import numpy

from transiflow import utils
from transiflow.Discretization import Discretization

class CylindricalDiscretization(Discretization):
    '''Finite volume discretization of the incompressible
    Navier-Stokes equations on a (possibly non-uniform) Arakawa C-grid
    in a cylindrical coordinate system. For details on the
    implementation and the ordering of the variables, see the
    :ref:`transiflow.discretization` class.

    '''

    def __init__(self, parameters, nr, ntheta, nz, dim=None, dof=None,
                 r=None, theta=None, z=None, boundary_conditions=None):
        self.parameters = parameters

        ri = self.parameters.get('R-min', 1.0)
        ro = self.parameters.get('R-max', 2.0)
        self.eta = ri / ro

        L = self.parameters.get('Z-max', 1.0) - self.parameters.get('Z-min', 0.0)

        r = self.get_coordinate_vector(1, 1 / self.eta, nr) if r is None else r

        theta = utils.create_uniform_coordinate_vector(self.parameters.get('Theta-min', 0.0),
                                                       self.parameters.get('Theta-max', 2 * numpy.pi),
                                                       ntheta) if theta is None else theta

        z = utils.create_uniform_coordinate_vector(0, L / self.eta, nz) if z is None else z

        Discretization.__init__(self, parameters, nr, ntheta, nz, dim, dof,
                                r, theta, z, boundary_conditions)

        self.y_periodic = True
        if self.parameters.get('Z-periodic', False):
            self.z_periodic = True

    def _linear_part_2D(self):
        '''Compute the linear part of the equation in case the domain is 2D.
        In case Re = 0 we instead compute the linear part for the Stokes
        problem.'''

        Re = self.get_parameter('Reynolds Number')

        if Re == 0:
            Re = 1

        return 1 / Re * (self.iruscale(self.u_rr()) + self.iru2scale(self.u_tt() - self.value_u() - 2 * self.v_t_u())
                         + self.irvscale(self.v_rr()) + self.irv2scale(self.v_tt() - self.value_v() + 2 * self.u_t_v())) \
            - (self.p_r() + self.irvscale(self.p_t())) \
            + self.div()

    def _linear_part_3D(self):
        '''Compute the linear part of the equation in case the domain is 3D.
        In case Re = 0 we instead compute the linear part for the Stokes
        problem.'''

        Ta = self.get_parameter('Taylor Number')
        if Ta == 0:
            Ta = self.get_parameter('Reynolds Number') / (1 / self.eta - 1)

        if Ta == 0:
            Ta = 1

        return 1 / Ta * (self.iruscale(self.u_rr()) + self.iru2scale(self.u_tt() - self.value_u() - 2 * self.v_t_u())
                         + self.u_zz()
                         + self.irvscale(self.v_rr()) + self.irv2scale(self.v_tt() - self.value_v() + 2 * self.u_t_v())
                         + self.v_zz()
                         + self.irvscale(self.w_rr()) + self.irv2scale(self.w_tt()) + self.w_zz()) \
            - (self.p_r() + self.irvscale(self.p_t()) + self.p_z()) \
            + self.div()

    def nonlinear_part(self, state):
        r'''Compute the nonlinear part of the equation. In case $\Re = 0$ this
        does nothing.

        :meta private:

        '''

        state_mtx = utils.create_padded_state_mtx(state, self.nx, self.ny, self.nz, self.dof,
                                                  self.x_periodic, self.y_periodic, self.z_periodic)

        atomJ = numpy.zeros((self.nx, self.ny, self.nz, self.dof, self.dof, 3, 3, 3))
        atomF = numpy.zeros((self.nx, self.ny, self.nz, self.dof, self.dof, 3, 3, 3))

        Ta = self.get_parameter('Taylor Number')
        Re = self.get_parameter('Reynolds Number')
        if Re == 0 and Ta == 0:
            return (atomJ, atomF)

        self.u_u_r(atomJ, atomF, state_mtx)
        self.u_v_r(atomJ, atomF, state_mtx)
        self.v_u_t(atomJ, atomF, state_mtx)
        self.v_v_t(atomJ, atomF, state_mtx)

        self.v_v(atomJ, atomF, state_mtx)
        self.u_v(atomJ, atomF, state_mtx)

        if self.dim > 2:
            self.u_w_r(atomJ, atomF, state_mtx)
            self.v_w_t(atomJ, atomF, state_mtx)
            self.w_u_z(atomJ, atomF, state_mtx)
            self.w_v_z(atomJ, atomF, state_mtx)
            self.w_w_z(atomJ, atomF, state_mtx)

        atomJ += atomF

        return (atomJ, atomF)

    def _taylor_couette(self, boundary_conditions, atom):
        '''Boundary conditions for the Taylor-Couette problem'''
        ri = self.parameters.get('R-min', 1.0)
        ro = self.parameters.get('R-max', 2.0)
        wo = self.get_parameter('Outer Angular Velocity', 0)
        wi = self.get_parameter('Inner Angular Velocity', 1)

        # This is not supported by the non-dimensionalization
        assert wo == 0

        boundary_conditions.moving_lid_east(atom, (wo * ro) / (wi * ri))
        boundary_conditions.moving_lid_west(atom, 1)

        if self.dim <= 2 or self.nz <= 1:
            return boundary_conditions.get_forcing()

        asym = self.get_parameter('Asymmetry Parameter')
        frc2 = numpy.zeros((self.nx, self.ny, self.nz, self.dof))
        frc2[self.nx-1, 0, :, 2] = asym * numpy.cos(self.z[0:self.nz] / self.z[self.nz-1] * numpy.pi)

        frc = boundary_conditions.get_forcing()
        frc += utils.create_state_vec(frc2, self.nx, self.ny, self.nz, self.dof)

        if not self.z_periodic:
            boundary_conditions.no_slip_top(atom)
            boundary_conditions.no_slip_bottom(atom)

        return frc

    def _setup_boundary_conditions(self):
        '''Setup boundary conditions for the currently defined problem type.'''
        if self.boundary_conditions:
            return
        elif self.problem_type_equals('Taylor-Couette'):
            self.boundary_conditions = self._taylor_couette
        else:
            raise Exception('Invalid problem type %s' % self.get_parameter('Problem Type'))

    # Below are all of the discretizations of separate parts of
    # equations that we can solve using FVM. This takes into account
    # non-uniform grids. New discretizations such as derivatives have
    # to be implemented in a similar way.

    def rvscale(self, atom):
        '''Scale atom by r at the location of v

        :meta private:

        '''
        for i in range(self.nx):
            atom[i, :, :, :, :, :, :, :] *= (self.x[i] + self.x[i-1]) / 2
        return atom

    def iruscale(self, atom):
        '''Scale atom by 1/r at the location of u

        :meta private:

        '''
        for i in range(self.nx):
            atom[i, :, :, :, :, :, :, :] /= self.x[i]
        return atom

    def irvscale(self, atom):
        '''Scale atom by 1/r at the location of v

        :meta private:

        '''
        for i in range(self.nx):
            atom[i, :, :, :, :, :, :, :] /= (self.x[i] + self.x[i-1]) / 2
        return atom

    def iru2scale(self, atom):
        '''Scale atom by 1/r^2 at the location of u

        :meta private:

        '''
        for i in range(self.nx):
            atom[i, :, :, :, :, :, :, :] /= self.x[i] * self.x[i]
        return atom

    def irv2scale(self, atom):
        '''Scale atom by 1/r^2 at the location of v

        :meta private:

        '''
        for i in range(self.nx):
            atom[i, :, :, :, :, :, :, :] /= (self.x[i] + self.x[i-1]) * (self.x[i] + self.x[i-1]) / 4
        return atom

    @staticmethod
    def _u_rr(atom, i, j, k, x, y, z):
        # distance between u[i] and u[i-1]
        dx = x[i] - x[i-1]
        rv = x[i-1] + dx / 2
        # distance between u[i+1] and u[i]
        dxp1 = x[i+1] - x[i]
        rvp1 = x[i] + dxp1 / 2
        # volume size in the y direction
        dy = y[j] - y[j-1]
        # volume size in the z direction
        dz = z[k] - z[k-1]

        # second order finite difference
        atom[0] = rv / dx * dy * dz
        atom[2] = rvp1 / dxp1 * dy * dz
        atom[1] = -atom[0] - atom[2]

    def u_rr(self):
        ''':meta private:'''
        atom = numpy.zeros((self.nx, self.ny, self.nz, self.dof, self.dof, 3, 3, 3))
        for i, j, k in numpy.ndindex(self.nx, self.ny, self.nz):
            self._u_rr(atom[i, j, k, 0, 0, :, 1, 1], i, j, k, self.x, self.y, self.z)
        return atom

    def v_tt(self):
        ''':meta private:'''
        return self.v_yy()

    @staticmethod
    def _v_rr(atom, i, j, k, x, y, z):
        # distance between v[i] and v[i-1]
        dx = (x[i] - x[i-2]) / 2
        # distance between v[i+1] and v[i]
        dxp1 = (x[i+1] - x[i-1]) / 2
        # volume size in the y direction
        dy = (y[j+1] - y[j-1]) / 2
        # volume size in the z direction
        dz = z[k] - z[k-1]

        # second order finite difference
        atom[0] = x[i-1] / dx * dy * dz
        atom[2] = x[i] / dxp1 * dy * dz
        atom[1] = -atom[0] - atom[2]

    def u_tt(self):
        ''':meta private:'''
        return self.u_yy()

    def v_rr(self):
        ''':meta private:'''
        atom = numpy.zeros((self.nx, self.ny, self.nz, self.dof, self.dof, 3, 3, 3))
        for i, j, k in numpy.ndindex(self.nx, self.ny, self.nz):
            self._v_rr(atom[i, j, k, 1, 1, :, 1, 1], i, j, k, self.x, self.y, self.z)
        return atom

    def w_tt(self):
        ''':meta private:'''
        return self.w_yy()

    def w_rr(self):
        ''':meta private:'''
        atom = numpy.zeros((self.nx, self.ny, self.nz, self.dof, self.dof, 3, 3, 3))
        for i, j, k in numpy.ndindex(self.nx, self.ny, self.nz):
            self._v_rr(atom[i, j, k, 2, 2, :, 1, 1], i, k, j, self.x, self.z, self.y)
        return atom

    def p_r(self):
        ''':meta private:'''
        return self.p_x()

    def p_t(self):
        ''':meta private:'''
        return self.p_y()

    def v_t_u(self):
        ''':meta private:'''
        atom = numpy.zeros((self.nx, self.ny, self.nz, self.dof, self.dof, 3, 3, 3))
        for i, j, k in numpy.ndindex(self.nx, self.ny, self.nz):
            Discretization._backward_u_y(atom[i, j, k, 0, 1, 1, :, 1], i, j, k, self.x, self.y, self.z)
            Discretization._backward_u_y(atom[i, j, k, 0, 1, 2, :, 1], i, j, k, self.x, self.y, self.z)
            atom[i, j, k, 0, 1, :, :, :] /= 2
        return atom

    def u_t_v(self):
        ''':meta private:'''
        atom = numpy.zeros((self.nx, self.ny, self.nz, self.dof, self.dof, 3, 3, 3))
        for i, j, k in numpy.ndindex(self.nx, self.ny, self.nz):
            Discretization._forward_u_x(atom[i, j, k, 1, 0, 0, :, 1], j, i, k, self.y, self.x, self.z)
            Discretization._forward_u_x(atom[i, j, k, 1, 0, 1, :, 1], j, i, k, self.y, self.x, self.z)
            atom[i, j, k, 1, 0, :, :, :] /= 2
        return atom

    @staticmethod
    def _value_u(atom, i, j, k, x, y, z):
        # volume size in the x direction
        dx = (x[i+1] - x[i-1]) / 2
        # volume size in the y direction
        dy = y[j] - y[j-1]
        # volume size in the z direction
        dz = z[k] - z[k-1]

        atom[1] = dx * dy * dz

    def value_u(self):
        ''':meta private:'''
        atom = numpy.zeros((self.nx, self.ny, self.nz, self.dof, self.dof, 3, 3, 3))
        for i, j, k in numpy.ndindex(self.nx, self.ny, self.nz):
            self._value_u(atom[i, j, k, 0, 0, :, 1, 1], i, j, k, self.x, self.y, self.z)
        return atom

    def value_v(self):
        ''':meta private:'''
        atom = numpy.zeros((self.nx, self.ny, self.nz, self.dof, self.dof, 3, 3, 3))
        for i, j, k in numpy.ndindex(self.nx, self.ny, self.nz):
            self._value_u(atom[i, j, k, 1, 1, :, 1, 1], j, i, k, self.y, self.x, self.z)
        return atom

    @staticmethod
    def _backward_u_r(atom, i, j, k, x, y, z):
        # volume size in the y direction
        dy = y[j] - y[j-1]
        # volume size in the z direction
        dz = z[k] - z[k-1]

        # backward difference
        atom[1] = x[i] * dy * dz
        atom[0] = -x[i-1] * dy * dz

    def u_r(self):
        ''':meta private:'''
        atom = numpy.zeros((self.nx, self.ny, self.nz, self.dof, self.dof, 3, 3, 3))
        for i, j, k in numpy.ndindex(self.nx, self.ny, self.nz):
            self._backward_u_r(atom[i, j, k, self.dim, 0, :, 1, 1], i, j, k, self.x, self.y, self.z)
        return atom

    def div(self):
        ''':meta private:'''
        if self.dim == 2:
            return self.u_x() + self.v_y()
        return self.u_x() + self.v_y() + self.rvscale(self.w_z())

    def u_u_r(self, atomJ, atomF, state):
        ''':meta private:'''
        Discretization.u_u_x(self, atomJ, atomF, state)

    def u_v_r(self, atomJ, atomF, state):
        ''':meta private:'''
        Discretization.u_v_x(self, atomJ, atomF, state)

    def u_w_r(self, atomJ, atomF, state):
        ''':meta private:'''
        Discretization.u_w_x(self, atomJ, atomF, state)

    def v_u_t(self, atomJ_in, atomF_in, state):
        ''':meta private:'''
        atomJ = numpy.zeros((self.nx, self.ny, self.nz, self.dof, self.dof, 3, 3, 3))
        atomF = numpy.zeros((self.nx, self.ny, self.nz, self.dof, self.dof, 3, 3, 3))

        Discretization.v_u_y(self, atomJ, atomF, state)
        self.iruscale(atomJ)
        self.iruscale(atomF)

        atomJ_in += atomJ
        atomF_in += atomF

    def v_v_t(self, atomJ_in, atomF_in, state):
        ''':meta private:'''
        atomJ = numpy.zeros((self.nx, self.ny, self.nz, self.dof, self.dof, 3, 3, 3))
        atomF = numpy.zeros((self.nx, self.ny, self.nz, self.dof, self.dof, 3, 3, 3))

        Discretization.v_v_y(self, atomJ, atomF, state)
        self.irvscale(atomJ)
        self.irvscale(atomF)

        atomJ_in += atomJ
        atomF_in += atomF

    def v_w_t(self, atomJ_in, atomF_in, state):
        ''':meta private:'''
        atomJ = numpy.zeros((self.nx, self.ny, self.nz, self.dof, self.dof, 3, 3, 3))
        atomF = numpy.zeros((self.nx, self.ny, self.nz, self.dof, self.dof, 3, 3, 3))

        Discretization.v_w_y(self, atomJ, atomF, state)
        self.irvscale(atomJ)
        self.irvscale(atomF)

        atomJ_in += atomJ
        atomF_in += atomF

    def v_v(self, atomJ_in, atomF_in, state):
        ''':meta private:'''
        averages_v = self.weighted_average_x(state[:, :, :, 1])
        averages_v = (averages_v[:, 0:self.ny, :] + averages_v[:, 1:self.ny+1, :]) / 2

        atom = numpy.zeros((self.nx, self.ny, self.nz, self.dof, self.dof, 3, 3, 3))

        atom_value = numpy.zeros(1)
        atom_average = numpy.zeros(2)
        for i, j, k in numpy.ndindex(self.nx, self.ny, self.nz):
            Discretization._mass_x(atom_value, i, j, k, self.x, self.y, self.z)
            Discretization._weighted_average(atom_average, i, self.x)
            atom[i, j, k, 0, 1, 1:3, 0, 1] += atom_value * atom_average * averages_v[i, j, k+1] * 1 / 2
            atom[i, j, k, 0, 1, 1:3, 1, 1] += atom_value * atom_average * averages_v[i, j, k+1] * 1 / 2

        self.iruscale(atom)

        atomJ_in += atom
        atomF_in += atom

    def u_v(self, atomJ_in, atomF_in, state):
        ''':meta private:'''
        averages_u = self.weighted_average_y(state[:, :, :, 0])
        averages_u = (averages_u[0:self.nx, :, :] + averages_u[1:self.nx+1, :, :]) / 2
        averages_v = state[1:self.nx+1, 1:self.ny+1, 1:self.nz+1, 1]

        atomJ = numpy.zeros((self.nx, self.ny, self.nz, self.dof, self.dof, 3, 3, 3))
        atomF = numpy.zeros((self.nx, self.ny, self.nz, self.dof, self.dof, 3, 3, 3))

        atom_value = numpy.zeros(1)
        atom_average = numpy.zeros(2)
        for i, j, k in numpy.ndindex(self.nx, self.ny, self.nz):
            Discretization._mass_x(atom_value, j, i, k, self.y, self.x, self.z)
            atomF[i, j, k, 1, 1, 1, 1:2, 1] -= atom_value * averages_u[i, j, k+1]

            Discretization._weighted_average(atom_average, j, self.y)
            atomJ[i, j, k, 1, 0, 0, 1:3, 1] -= atom_value * atom_average * averages_v[i, j, k] * 1 / 2
            atomJ[i, j, k, 1, 0, 1, 1:3, 1] -= atom_value * atom_average * averages_v[i, j, k] * 1 / 2

        self.irvscale(atomF)
        self.irvscale(atomJ)

        atomJ_in += atomJ
        atomF_in += atomF



from PyTrilinos import Epetra
from PyTrilinos import Amesos

from transiflow.interface import ParallelBaseInterface


class Vector(Epetra.Vector):
    '''Distributed Epetra_Vector with some extra methods added to it for convenience.'''
    def __neg__(self):
        v = Vector(self)
        v.Scale(-1.0)
        return v

    def __truediv__(self, scal):
        v = Vector(self)
        v.Scale(1.0 / scal)
        return v

    def dot(self, other):
        return self.Dot(other)[0]

    def gather(self):
        local_elements = []
        if self.Comm().MyPID() == 0:
            local_elements = range(self.Map().NumGlobalElements())

        local_map = Epetra.Map(-1, local_elements, 0, self.Comm())
        importer = Epetra.Import(local_map, self.Map())
        out = Vector(local_map)
        out.Import(self, importer, Epetra.Insert)
        return out

    def all_gather(self):
        local_elements = range(self.Map().NumGlobalElements())
        local_map = Epetra.Map(-1, local_elements, 0, self.Comm())
        importer = Epetra.Import(local_map, self.Map())
        out = Vector(local_map)
        out.Import(self, importer, Epetra.Insert)
        return out

    @staticmethod
    def from_array(m, x):
        local_elements = []
        if m.Comm().MyPID() == 0:
            local_elements = range(m.NumGlobalElements())

        local_map = Epetra.Map(-1, local_elements, 0, m.Comm())
        importer = Epetra.Import(m, local_map)
        x_local = Vector(Epetra.Copy, local_map, x)
        out = Vector(m)
        out.Import(x_local, importer, Epetra.Insert)
        return out

    size = property(Epetra.Vector.GlobalLength)


class Interface(ParallelBaseInterface):
    '''This class defines an interface to the Epetra backend for the
    discretization. This backend can be used for testing
    low-resolution parallel runs as well as a base for a class that
    implements a better linear solver. It should not be used for any
    meaningful simulations.

    The Epetra backend partitions the domain into Cartesian subdomains. The
    subdomains will be distributed over multiple processors if MPI is
    used to run the application.

    See :mod:`.Discretization` for the descriptions of the constructor
    arguments.

    Parameters
    ----------
    parameters : dict
        Key-value pairs that can be used to modify parameters in the
        discretization as well as the linear solver and eigenvalue
        solver.

    '''
    def __init__(self, parameters, nx, ny, nz=1, dim=None, dof=None,
                 boundary_conditions=None, comm=None):
        if comm is None:
            comm = Epetra.PyComm()

        ParallelBaseInterface.__init__(self, comm, parameters, nx, ny, nz, dim, dof,
                                       boundary_conditions)

        self.map = self.create_map()

        self.assembly_map = self.create_map(True)
        self.assembly_importer = Epetra.Import(self.assembly_map, self.map)

        self.solve_map = self.map
        self.solve_importer = Epetra.Import(self.solve_map, self.map)

        self.left_scaling = None
        self.inv_left_scaling = None
        self.right_scaling = None
        self.inv_right_scaling = None

        self.jac = None
        self.mass = None

    def vector(self, *args):
        if len(args) == 0:
            return Vector(self.map)

        return Vector(args)

    def vector_from_array(self, array):
        return Vector.from_array(self.map, array)

    def array_from_vector(self, vector):
        return Vector.all_gather(vector)

    def create_map(self, overlapping=False):
        '''Create a map on which the local discretization domain is defined.
        The overlapping part is only used for computing the discretization.

        :meta private:

        '''

        local_elements = ParallelBaseInterface.create_map(self, overlapping)
        return Epetra.Map(-1, local_elements, 0, self.comm)

    def rhs(self, state):
        r'''Compute the right-hand side of the DAE. That is the
        right-hand side $F(u, p)$ in

        .. math:: M(p) \frac{\d u}{\d t} = F(u, p)

        The RHS is computed on the overlapping subdomain map and then
        distributed using the non-overlapping map that is used by the
        linear solver.

        Parameters
        ----------
        state : array_like
            State $u$ at which to evaluate $F(u, p)$.

        Returns
        -------
        rhs : array_like
            The value of $F(u, p)$.

        '''
        state_ass = Vector(self.assembly_map)
        state_ass.Import(state, self.assembly_importer, Epetra.Insert)

        rhs = self.discretization.rhs(state_ass)
        rhs_ass = Vector(Epetra.Copy, self.assembly_map, rhs)
        rhs = Vector(self.map)
        rhs.Export(rhs_ass, self.assembly_importer, Epetra.Zero)
        return rhs

    def jacobian(self, state):
        r'''Compute the Jacobian matrix $J(u, p)$ of the right-hand
        side of the DAE. That is the Jacobian matrix of $F(u, p)$ in

        .. math:: M(p) \frac{\d u}{\d t} = F(u, p)

        The Jacobian matrix is computed on the overlapping subdomain
        map and then distributed using the non-overlapping map that is
        used by the linear solver.

        Parameters
        ----------
        state : array_like
            State $u$ at which to evaluate $J(u, p)$.

        Returns
        -------
        jac : Epetra.FECrsMatrix
            The matrix $J(u, p)$ in a CSR matrix format that allows
            for distributing the overlap.

        '''
        state_ass = Vector(self.assembly_map)
        state_ass.Import(state, self.assembly_importer, Epetra.Insert)

        local_jac = self.discretization.jacobian(state_ass)

        if self.jac is None:
            self.jac = Epetra.FECrsMatrix(Epetra.Copy, self.solve_map, 27)
        else:
            self.jac.PutScalar(0.0)

        for i in range(len(local_jac.begA) - 1):
            if self.is_ghost(i):
                continue
            row = self.assembly_map.GID64(i)
            for j in range(local_jac.begA[i], local_jac.begA[i + 1]):
                # __setitem__ automatically calls ReplaceGlobalValues if the matrix is filled,
                # InsertGlobalValues otherwise
                self.jac[row, self.assembly_map.GID64(local_jac.jcoA[j])] = local_jac.coA[j]
        self.jac.GlobalAssemble(True, Epetra.Insert)

        return self.jac

    def mass_matrix(self):
        r'''Compute the mass matrix of the DAE. That is the mass matrix
        $M(p)$ in

        .. math:: M(p) \frac{\d u}{\d t} = F(u, p)

        The mass matrix is computed on the overlapping subdomain map
        and then distributed using the non-overlapping map that is
        used by the linear solver.

        Returns
        -------
        mass : Epetra.FECrsMatrix
            The matrix $M(p)$ in a CSR matrix format that allows for
            distributing the overlap.

        '''
        local_mass = self.discretization.mass_matrix()

        if self.mass is None:
            self.mass = Epetra.FECrsMatrix(Epetra.Copy, self.solve_map, 1)
        else:
            self.mass.PutScalar(0.0)

        for i in range(len(local_mass.begA) - 1):
            if self.is_ghost(i):
                continue
            row = self.assembly_map.GID64(i)
            for j in range(local_mass.begA[i], local_mass.begA[i + 1]):
                # __setitem__ automatically calls ReplaceGlobalValues if the matrix is filled,
                # InsertGlobalValues otherwise
                self.mass[row, self.assembly_map.GID64(local_mass.jcoA[j])] = local_mass.coA[j]
        self.mass.GlobalAssemble(True, Epetra.Insert)

        return self.mass

    def _compute_scaling(self):
        '''Compute scaling for the linear problem. This scaling makes
        solving the linear system more stable.'''

        self.left_scaling = Vector(self.solve_map)
        self.left_scaling.PutScalar(1.0)

        self.inv_left_scaling = Vector(self.solve_map)
        self.inv_left_scaling.PutScalar(1.0)

        self.right_scaling = Vector(self.solve_map)
        self.right_scaling.PutScalar(1.0)

        self.inv_right_scaling = Vector(self.solve_map)
        self.inv_right_scaling.PutScalar(1.0)

        dim = self.discretization.dim
        dof = self.discretization.dof

        for lrid in range(self.jac.NumMyRows()):
            grid = self.jac.GRID(lrid)
            var = grid % dof
            values, indices = self.jac.ExtractMyRowCopy(lrid)
            for j in range(len(indices)):
                lcid = indices[j]
                value = values[j]
                gcid = self.jac.GCID(lcid)
                if value < 1e-8:
                    continue

                if var != dim and gcid % dof == dim:
                    # If the row is a velocity and the column a pressure
                    self.left_scaling[lrid] = 1 / value
                    self.inv_left_scaling[lrid] = value
                    break
                elif var == dim and gcid % dof != dim and self.jac.MyGRID(gcid):
                    # If the row is a pressure and the column a velocity
                    lid = self.jac.LRID(gcid)
                    self.right_scaling[lid] = 1 / value
                    self.inv_right_scaling[lid] = value

    def _scale_matrix(self, mat):
        assert not hasattr(mat, 'scaled') or not mat.scaled
        mat.scaled = True

        mat.LeftScale(self.left_scaling)
        mat.RightScale(self.right_scaling)

    def _scale_jacobian(self):
        self._scale_matrix(self.jac)

    def _scale_rhs(self, rhs):
        rhs.Multiply(1.0, self.left_scaling, rhs, 0.0)

    def _scale_lhs(self, lhs):
        lhs.Multiply(1.0, self.inv_right_scaling, lhs, 0.0)

    def _unscale_matrix(self, mat):
        assert mat.scaled
        mat.scaled = False

        mat.LeftScale(self.inv_left_scaling)
        mat.RightScale(self.inv_right_scaling)

    def _unscale_jacobian(self):
        self._unscale_matrix(self.jac)

    def _unscale_rhs(self, rhs):
        rhs.Multiply(1.0, self.inv_left_scaling, rhs, 0.0)

    def _unscale_lhs(self, lhs):
        lhs.Multiply(1.0, self.right_scaling, lhs, 0.0)

    def solve(self, jac, rhs, rhs2=None, V=None, W=None, C=None, solver=None):
        if rhs2:
            raise NotImplementedError()

        x = Vector(rhs)

        problem = Epetra.LinearProblem(jac, x, rhs)
        factory = Amesos.Factory()
        solver = factory.Create('Klu', problem)
        solver.SymbolicFactorization()
        solver.NumericFactorization()
        solver.Solve()

        return x

    def eigs(self, state, return_eigenvectors=False, enable_recycling=False):
        raise NotImplementedError()



import math
import numpy

from transiflow.utils import norm

class Continuation:
    '''Pseudo-arclength continuation of an abstract interface that can
    return a Jacobian matrix and a right-hand side.

    Parameters
    ----------
    interface
        Interface object that implements the following functions:
        ``rhs(x)``, ``jacobian(x)``, ``solve(jac, rhs)``, ``eigs(x)``
        and ``set_parameter(name, value)``.
    delta : scalar, optional
        Parameter value difference used when computing the approximate
        derivative.
    destination_tolerance : scalar, optional
        Tolerance used for the continuation parameter when converging
        onto a bifurcation point.
    newton_tolerance : scalar, optional
        Tolerance used in the Newton corrector to determine
        convergence.
    maximum_newton_iterations : int, optional
        Maximum number of Newton iterations.
    optimal_newton_iterations : int, optional
        Number of Newton iterations in which we try to converge by
        adjusting ds.
    residual_check: str, optional
        Method for checking the residual in the Newton method
        (default: 'F').

        - ``F``: Use the norm of the ``rhs(x)`` function.
        - ``dx``: Use the norm of the step ``dx``.
    bordered_solver : bool, optional
        Use a bordered solver the entire linear system in the
        correction equation in a single operation, instead of using
        two separate solves. This has to be implemented by the
        interface.
    verbose : bool, optional
        Give extra information about convergence. Since we have to
        compute this information, this may be slower.

    '''

    def __init__(self, interface,
                 delta=1.0, destination_tolerance=1e-4,
                 newton_tolerance=1e-4, maximum_newton_iterations=10,
                 optimal_newton_iterations=3, residual_check='F',
                 bordered_solver=False, verbose=False):
        self.interface = interface
        self.verbose = verbose

        self.destination_tolerance = destination_tolerance
        self.delta = delta
        self.zeta = None

        self.maximum_newton_iterations = maximum_newton_iterations
        self.optimal_newton_iterations = optimal_newton_iterations
        self.newton_tolerance = newton_tolerance
        self.newton_iterations = 0
        self.residual_check = residual_check

        self.bordered_solver = bordered_solver

    def newton(self, x0, tol=1e-10):
        '''Newton method that can be used to converge to a steady
        state that is very nearby. This can for instance be used to
        obtain an initial guess.

        Parameters
        ----------
        x0 : array_like
            Initial solution.
        tol : scalar, optional
            Convergence tolerance

        Returns
        -------
        x : array_like
            Root at the current parameter value.

        '''

        x = x0
        for k in range(self.maximum_newton_iterations):
            fval = self.interface.rhs(x)

            if self.residual_check == 'F' or self.verbose:
                fnorm = norm(fval)

            if self.residual_check == 'F' and fnorm < tol:
                print('Newton converged in %d iterations with ||F||=%e' % (k, fnorm), flush=True)
                break

            jac = self.interface.jacobian(x)
            dx = self.interface.solve(jac, -fval)

            x = x + dx

            if self.residual_check != 'F' or self.verbose:
                dxnorm = norm(dx)

            if self.residual_check != 'F' and dxnorm < tol:
                print('Newton converged in %d iterations with ||dx||=%e' % (k, dxnorm), flush=True)
                break

            if self.verbose:
                print('Newton status at iteration %d: ||F||=%e, ||dx||=%e' % (k, fnorm, dxnorm), flush=True)

        self.newton_iterations = k

        return x

    def _newton_corrector(self, parameter_name, ds, x, x0, mu, mu0):
        self.newton_iterations = 0

        fnorm = None
        dxnorm = None

        # Do the main iteration
        for k in range(self.maximum_newton_iterations):
            # Compute F (RHS of 2.2.9)
            self.interface.set_parameter(parameter_name, mu)
            fval = self.interface.rhs(x)

            if self.residual_check == 'F' or self.verbose:
                prev_norm = fnorm
                fnorm = norm(fval)

            if self.residual_check == 'F' and fnorm < self.newton_tolerance:
                print('Newton corrector converged in %d iterations with ||F||=%e' % (k, fnorm), flush=True)
                break

            if self.residual_check == 'F' and prev_norm is not None and prev_norm < fnorm:
                self.newton_iterations = self.maximum_newton_iterations
                break

            # Compute r (2.2.8)
            diff = x - x0
            rnp1 = self.zeta * diff.dot(diff) + (1 - self.zeta) * (mu - mu0) ** 2 - ds ** 2

            # Compute F_mu (LHS of 2.2.9)
            self.interface.set_parameter(parameter_name, mu + self.delta)
            dflval = (self.interface.rhs(x) - fval) / self.delta
            self.interface.set_parameter(parameter_name, mu)

            # Compute the jacobian F_x at x (LHS of 2.2.9)
            jac = self.interface.jacobian(x)

            if self.bordered_solver:
                # Solve the entire bordered system in one go (2.2.9)
                dx, dmu = self.interface.solve(jac, -fval, -rnp1, dflval, 2 * self.zeta * diff,
                                               2 * (1 - self.zeta) * (mu - mu0))
            else:
                # Solve twice with F_x (2.2.9)
                z1 = self.interface.solve(jac, -fval)
                z2 = self.interface.solve(jac, dflval)

                # Compute dmu (2.2.13)
                dmu = (-rnp1 - 2 * self.zeta * diff.dot(z1)) / (2 * (1 - self.zeta) * (mu - mu0)
                                                                - 2 * self.zeta * diff.dot(z2))

                # Compute dx (2.2.12)
                dx = z1 - dmu * z2

            # Compute a new x and mu (2.2.10 - 2.2.11)
            x += dx
            mu += dmu

            self.newton_iterations += 1

            if self.residual_check != 'F' or self.verbose:
                prev_norm = dxnorm
                dxnorm = norm(dx)

            if self.residual_check != 'F' and dxnorm < self.newton_tolerance:
                print('Newton corrector converged in %d iterations with ||dx||=%e' % (k, dxnorm), flush=True)
                break

            if self.verbose:
                print('Newton corrector status at iteration %d: ||F||=%e, ||dx||=%e' % (k, fnorm, dxnorm), flush=True)

            if self.residual_check != 'F' and prev_norm is not None and prev_norm < dxnorm:
                self.newton_iterations = self.maximum_newton_iterations
                break

        if self.newton_iterations == self.maximum_newton_iterations:
            print('Newton did not converge. Adjusting step size and trying again', flush=True)
            return x0, mu0

        self.interface.set_parameter(parameter_name, mu)

        return x, mu

    def _adjust_step_size(self, ds, ds_min, ds_max):
        ''' Step size control, see [Seydel p 188.] '''
        factor = self.optimal_newton_iterations / max(self.newton_iterations, 1)
        factor = min(max(factor, 0.5), 2.0)

        ds *= factor

        ds = math.copysign(min(max(abs(ds), ds_min), ds_max), ds)

        if self.verbose:
            print('New stepsize: ds=%e, factor=%e' % (ds, factor), flush=True)

        return ds

    def _detect_bifurcation(self, parameter_name, x, mu, dx, dmu, eigs, deig, v,
                            ds, ds_min, ds_max, maxit):
        ''' Converge onto a bifurcation '''

        for j in range(maxit):
            i = numpy.argmin(numpy.abs(eigs.real))
            if abs(eigs[i].real) < self.destination_tolerance:
                print("Bifurcation found at %s = %f with eigenvalue %e + %ei" % (
                    parameter_name, mu, eigs[i].real, eigs[i].imag), flush=True)
                break

            # Secant method
            ds = ds / deig.real * -eigs.real[i]
            x, mu, dx, dmu, ds = self._step(parameter_name, x, mu, dx, dmu,
                                            ds, ds_min, ds_max)

            prev_eigs = eigs
            eigs, v = self.interface.eigs(x, return_eigenvectors=True, enable_recycling=True)
            i = numpy.argmin(numpy.abs(eigs.real))
            deig = eigs[i] - prev_eigs[i]

        return x, mu, v

    def _converge(self, parameter_name, x, mu, dx, dmu, target,
                  ds, ds_min, ds_max, maxit):
        ''' Converge onto the target value '''

        for j in range(maxit):
            if abs(target - mu) < self.destination_tolerance:
                self.interface.set_parameter(parameter_name, target)
                print("Convergence achieved onto target %s = %f" % (parameter_name, mu), flush=True)
                break

            # Secant method
            ds = 1 / dmu * (target - mu)
            x, mu, dx, dmu, ds = self._step(parameter_name, x, mu, dx, dmu,
                                            ds, ds_min, ds_max)

        return x, mu

    def _step(self, parameter_name, x, mu, dx, dmu,
              ds, ds_min, ds_max):
        ''' Perform one step of the continuation '''

        mu0 = mu
        x0 = x

        # Predictor (2.2.3)
        mu = mu0 + ds * dmu
        x = x0 + ds * dx

        # Corrector (2.2.9 and onward)
        x, mu = self._newton_corrector(parameter_name, ds, x, x0, mu, mu0)

        if mu == mu0:
            # No convergence was achieved, adjusting the step size
            prev_ds = ds
            ds = self._adjust_step_size(ds, ds_min, ds_max)
            if prev_ds == ds:
                raise Exception('Newton cannot achieve convergence')

            return self._step(parameter_name, x0, mu0, dx, dmu,
                              ds, ds_min, ds_max)

        print("%s: %f" % (parameter_name, mu), flush=True)

        # Set the new values computed by the corrector
        dmu = mu - mu0
        dx = x - x0

        if abs(dmu) < 1e-12:
            raise Exception('dmu too small')

        # Compute the tangent (2.2.4)
        dx /= ds
        dmu /= ds

        return x, mu, dx, dmu, ds

    def _switch_branches_tangent(self, parameter_name, x, mu, dx, dmu, v,
                                 ds, ds_min, ds_max):
        ''' Switch branches according to (5.16) '''

        dmu0 = dmu

        # Compute the F(x) and F_x(x)
        fval = self.interface.rhs(x)
        jac = self.interface.jacobian(x)

        # Compute F_mu(x)
        self.interface.set_parameter(parameter_name, mu + self.delta)
        dflval = (self.interface.rhs(x) - fval) / self.delta
        self.interface.set_parameter(parameter_name, mu)

        if self.bordered_solver:
            # Solve the entire bordered system in one go (5.16)
            dx, dmu = self.interface.solve(jac, 0 * x, 0, dflval, dx, dmu)
        else:
            # Solve with F_x (equation below 5.16)
            z = self.interface.solve(jac, dflval)
            dmu = -dx.dot(v) / (dmu - dx.dot(z))
            dx = v - dmu * z

        if abs(dmu) < 1e-12:
            dmu = dmu0
            ds = ds_min

        return x, mu, dx, dmu, ds

    def _switch_branches(self, parameter_name, x, mu, dx, dmu, v,
                         ds, ds_min, ds_max):
        return self._switch_branches_tangent(parameter_name, x, mu, dx, dmu, v,
                                             ds, ds_min, ds_max)

    def _num_positive_eigs(self, eigs):
        # Include the range of the destination tolerance here to make sure
        # we don't converge onto the same target twice
        return sum([eig.real > -self.destination_tolerance for eig in eigs])

    def _initial_tangent(self, x, parameter_name, mu):
        ''' Compute the initial tangent '''

        # Get the initial tangent (2.2.5 - 2.2.7).
        self.interface.set_parameter(parameter_name, mu + self.delta)
        dflval = self.interface.rhs(x)
        self.interface.set_parameter(parameter_name, mu)
        fval = self.interface.rhs(x)
        dflval = (dflval - fval) / self.delta

        # Compute the jacobian at x and solve with it (2.2.5)
        jac = self.interface.jacobian(x)
        dx = self.interface.solve(jac, -dflval)

        # Scaling of the initial tangent (2.2.7)
        dmu = 1
        nrm = math.sqrt(self.zeta * dx.dot(dx) + dmu ** 2)
        dmu /= nrm
        dx /= nrm

        return dx, dmu

    def _return(self, x, mu, dx, dmu, ds, callback, return_step):
        if callback is not None:
            callback(self.interface, x, mu)

        if return_step:
            return x, mu, dx * ds, dmu * ds

        return x, mu

    def continuation(self, x0, parameter_name, start, target,
                     ds, ds_min=0.01, ds_max=1000,
                     dx=None, dmu=None, maxit=1000,
                     detect_bifurcations=False, switch_branches=False,
                     return_step=False, callback=None):
        '''Perform a pseudo-arclength continuation in
        ``parameter_name`` from parameter value start to ``target``
        with arclength step size ``ds``, and starting from an initial
        state ``x0``.

        Returns the final state x and the final parameter value mu.

        Parameters
        ----------
        x0 : array_like
            Initial solution.
        parameter_name : string
            Name of the parameter we want to perform the continuation
            in.
        start : scalar
            Starting value of the continuation parameter.
        target : scalar
            Target value of the continuation parameter.
        ds : scalar
            Arclength step size.
        ds_min : scalar, optional
            Minimum arclength step size.
        ds_max : scalar, optional
            Maximum arclength step size.
        dx : array_like, optional
            Vector difference defining the initial tangent.
        dmu : scalar, optional
            Parameter difference defining the initial tangent.
        maxit : int, optional
            Maximum number of continuation iterations.
        detect_bifurcations : bool, optional
            Detect bifurcations by detecting a switch in eigenvalue
            signs. Note that this is very expensive.
        switch_branches : bool, optional
            Switch branches using the tangent. This usually doesn't
            work and is currently untested.
        return_step : bool, optional
            Return ``dx`` and ``dmu`` when set to True. These can be
            used in the next ``continuation()`` call.
        callback : function, optional
            User-supplied function to call after each continuation
            step. It is called as ``callback(interface, x, mu)``.

        Returns
        -------
        x : array_like
            Value at the target or bifurcation point if were are
            detecting bifurcation points.
        mu : scalar
            Value of the bifurcation parameter at the end of the
            continuation.
        dx : array_like, optional
            Array to pass back to ``continuation()`` which is returned
            if ``return_step`` is set to True.
        dmu : scalar, optional
            Value to pass back to ``continuation()`` which is returned
            if ``return_step`` is set to True.

        '''

        x = x0
        mu = start

        self.zeta = 1 / x.size

        if dx is None or dmu is None:
            # Get the initial tangent (2.2.5 - 2.2.7).
            dx, dmu = self._initial_tangent(x, parameter_name, mu)
        else:
            dx /= ds
            dmu /= ds

        switched_branches = False
        eigs = None

        # Some configuration for the detection of bifurcations
        enable_recycling = False

        # Perform the continuation
        for j in range(maxit):
            mu0 = mu

            if detect_bifurcations or (switch_branches and not switched_branches):
                prev_eigs = eigs
                eigs, v = self.interface.eigs(x, return_eigenvectors=True, enable_recycling=enable_recycling)
                enable_recycling = True

                if prev_eigs is not None and self._num_positive_eigs(eigs) != self._num_positive_eigs(prev_eigs):
                    i = numpy.argmin(numpy.abs(eigs.real))
                    deig = eigs[i] - prev_eigs[i]
                    x, mu, v = self._detect_bifurcation(parameter_name, x, mu, dx, dmu, eigs, deig, v,
                                                        ds, ds_min, ds_max, maxit - j)

                    if switch_branches and not switched_branches:
                        switched_branches = True
                        x, mu, dx, dmu, ds = self._switch_branches(parameter_name, x, mu, dx, dmu, v[:, 0].real,
                                                                   ds, ds_min, ds_max)
                        continue

                    return self._return(x, mu, dx, dmu, ds, callback, return_step)

            if callback is not None:
                callback(self.interface, x, mu)

            x, mu, dx, dmu, ds = self._step(parameter_name, x, mu, dx, dmu,
                                            ds, ds_min, ds_max)

            if (mu >= target and mu0 < target) or (mu <= target and mu0 > target):
                # Converge onto the end point
                x, mu = self._converge(parameter_name, x, mu, dx, dmu, target,
                                       ds, ds_min, ds_max, maxit - j)

                return self._return(x, mu, dx, dmu, ds, callback, return_step)

            ds = self._adjust_step_size(ds, ds_min, ds_max)

        return self._return(x, mu, dx, dmu, ds, callback, return_step)



import numpy
import matplotlib.pyplot as plt

from transiflow import utils
from transiflow.utils import create_state_mtx # noqa: F401

def get_meshgrid(interface, x=None, y=None):
    '''Wrapper around `numpy.meshgrid(x, y)` that obtains necessary
    information from the interface.

    Parameters
    ----------
    interface : Interface
        Interface containing the coordinate vectors.
    x : array_like, optional
        Override the `x` meshgrid argument.
    y : array_like, optional
        Override the `y` meshgrid argument.

    Returns
    -------
    x : array_like
        2D array containing x coordinates.
    y : array_like
        2D array containing y coordinates.

    '''
    if x is None:
        x = interface.x[:-3]
    if y is None:
        if interface.ny > 1:
            y = interface.y[:-3]
        else:
            y = interface.z[:-3]

    return numpy.meshgrid(x, y)

def plot_contour(x, y, value, axis=2, title=None, legend=True, grid=True,
                 show=True, color=True, labels=True, levels=15, inline=False):
    '''Helper for plotting a contour plot.

    Parameters
    ----------
    x : array_like
        2D array containing x coordinates.
    y : array_like
        2D array containing y coordinates.
    value : array_like
        2D array of the value to plot.
    axis : int, optional
        Axis to ignore. Used for axis labels.
    title : str, optional
        Title of the plot.
    legend : bool, optional
        Whether to add a colorbar.
    grid : bool, optional
        Whether to show the mesh.
    show : bool, optional
        Whether to show the plot. This can be disabled when using
        `savefig()` manually.
    color : bool, optional
        Can be set to False to make plots suitable for black and white
        printing.
    labels : bool, optional
        Whether to add labels to the axis.
    levels : int, optional
        Number of levels used for the contours.
    inline : bool, optional
        Add inline labels to the contours. Useful for black and white
        plots.

    Returns
    -------
    fig : Figure
        Figure object that can be used to make manual modifications to
        the plot after calling this function.

    '''
    fig, ax = plt.subplots()

    if color:
        cs = ax.contourf(x, y, value.transpose(), levels)
    else:
        cs = ax.contour(x, y, value.transpose(), levels, colors=['k'])

    if inline:
        ax.clabel(cs, cs.levels, inline=True, fontsize=10)

    if legend:
        fig.colorbar(cs)

    if grid:
        ax.vlines(x[0, :], *y[[0, -1], 0], colors='0.3', linewidths=0.5)
        ax.hlines(y[:, 0], *x[0, [0, -1]], colors='0.3', linewidths=0.5)

    if labels:
        if axis == 0:
            ax.set_xlabel('y')
        else:
            ax.set_xlabel('x')

        if axis == 2:
            ax.set_ylabel('y')
        else:
            ax.set_ylabel('z')

    if title:
        plt.title(title)

    if show:
        plt.show()

    return fig

def plot_velocity_magnitude(state, interface, axis=2, position=None, title='Velocity magnitude', *args, **kwargs):
    '''Create a plot of the velocity magnitude.

    See :meth:`plot_contour` and :func:`.compute_velocity_magnitude` for details.

    '''
    m = utils.compute_velocity_magnitude(state, interface, axis, position)

    x, y = get_meshgrid(interface)

    return plot_contour(x, y, m, axis=axis, title=title, *args, **kwargs)

def plot_streamfunction(state, interface, axis=2, title='Stream function', *args, **kwargs):
    '''Create a plot of the stream function.

    See :meth:`plot_contour` and :func:`.compute_streamfunction` for details.

    '''
    psi = utils.compute_streamfunction(state, interface, axis)

    x, y = get_meshgrid(interface)

    return plot_contour(x, y, psi, axis=axis, title=title, *args, **kwargs)

def plot_vorticity(state, interface, axis=2, title='Vorticity', *args, **kwargs):
    '''Create a plot of the vorticity.

    See :meth:`plot_contour` and :func:`.compute_vorticity` for details.

    '''
    psi = utils.compute_vorticity(state, interface, axis)

    x, y = get_meshgrid(interface)

    return plot_contour(x, y, psi, axis=axis, title=title, *args, **kwargs)

def plot_value(value, interface=None, x=None, y=None, title=None, *args, **kwargs):
    '''Create a plot of the velocity magnitude.

    See :meth:`plot_contour` for details and extra parameters.

    Parameters
    ----------
    value : array_like
        2D array of the value to plot.
    interface : Interface, optional
        Interface containing the coordinate vectors.
    x : array_like, optional
        First coordinate vector.
    y : array_like, optional
        Second coordinate vector.

    '''
    x, y = get_meshgrid(interface, x, y)

    return plot_contour(x, y, value, title=title, *args, **kwargs)



def _icmp(first, second):
    return first.lower() == second.lower()

def create(parameters, nx, ny, nz=1, dim=None, dof=None,
           boundary_conditions=None, backend='SciPy'):
    '''Helper function to create an interface with a certain backend.
    This functions is aliased as ``Interface``. It can be called as

    .. code-block:: Python

        from transiflow import Interface
        interface = Interface(parameters, nx, ny)

    Parameters
    ----------
    parameters : dict
        Key-value pairs describing parameters of the model, for
        instance the Renolds number and the problem type. Possible
        values can be found in :ref:`problem definitions`.
    nx : int
        Grid size in the x direction.
    ny : int
        Grid size in the y direction.
    nz : int, optional
        Grid size in the z direction. 1 for 2-dimensional problems.
        This is the default.
    dim : int, optional
        Physical dimension of the problem. In case this is set to 2, w
        is not referenced in the state vector. The default is based on
        the value of nz.
    dof : int, optional
        Degrees of freedom for this problem. This should be set to dim
        plus 1 for each of pressure, temperature and salinity, if they
        are required for your problem. For example a 3D differentially
        heated cavity has dof = 3 + 1 + 1 = 5.
    boundary_conditions : function, optional
        User-supplied function that implements the boundary
        conditions. It is called as ``boundary_conditions(bc, atom)``
        where ``bc`` is an instance of the :class:`.BoundaryConditions`
        class.
    backend : str, optional
        The backend to use. Can be ``Epetra``, ``HYMLS``, ``PETSc``, ``SciPy``.

    Returns
    -------
    interface : Interface
        An interface instance.

    '''

    if _icmp(backend, 'Epetra'):
        from .Epetra import Interface
        return Interface(parameters, nx, ny, nz, dim, dof,
                         boundary_conditions=boundary_conditions)

    if _icmp(backend, 'HYMLS'):
        from .HYMLS import Interface
        return Interface(parameters, nx, ny, nz, dim, dof,
                         boundary_conditions=boundary_conditions)

    if _icmp(backend, 'PETSc'):
        from .PETSc import Interface
        return Interface(parameters, nx, ny, nz, dim, dof,
                         boundary_conditions=boundary_conditions)

    from .SciPy import Interface
    return Interface(parameters, nx, ny, nz, dim, dof,
                     boundary_conditions=boundary_conditions)



import numpy
import pytest

from transiflow import TimeIntegration
from transiflow import Continuation
from transiflow import utils

from transiflow.interface import create


def Interface(*args, backend="SciPy", **kwargs):
    try:
        return create(*args, backend=backend, **kwargs)
    except ImportError:
        pytest.skip(backend + " not found")


@pytest.mark.parametrize("backend", ["SciPy", "Epetra", "HYMLS", "PETSc"])
def test_continuation_ldc(backend, nx=4):
    numpy.random.seed(1234)

    ny = nx
    nz = nx

    parameters = {}
    interface = Interface(parameters, nx, ny, nz, backend=backend)
    continuation = Continuation(interface)

    x0 = interface.vector()
    x0 = continuation.newton(x0)

    start = 0
    target = 100
    ds = 100
    x = continuation.continuation(x0, 'Reynolds Number', start, target, ds)[0]

    assert utils.norm(x) > 0

def continuation_ldc_semi_2D(backend, nx=4):
    numpy.random.seed(1234)

    dim = 3
    ny = nx
    nz = 1

    parameters = {}
    interface = Interface(parameters, nx, ny, nz, dim, backend=backend)
    continuation = Continuation(interface)

    x0 = interface.vector()
    x0 = continuation.newton(x0)

    start = 0
    target = 2000
    ds = 100
    x = continuation.continuation(x0, 'Reynolds Number', start, target, ds)[0]

    assert utils.norm(x) > 0
    return interface.array_from_vector(x)

def continuation_ldc_2D(backend, nx=4):
    numpy.random.seed(1234)

    dim = 2
    ny = nx
    nz = 1

    parameters = {}
    interface = Interface(parameters, nx, ny, nz, dim, backend=backend)
    continuation = Continuation(interface)

    x0 = interface.vector()
    x0 = continuation.newton(x0)

    start = 0
    target = 2000
    ds = 100
    x = continuation.continuation(x0, 'Reynolds Number', start, target, ds)[0]

    assert utils.norm(x) > 0
    return interface.array_from_vector(x)

@pytest.mark.parametrize("backend", ["SciPy", "HYMLS", "PETSc"])
def test_continuation_ldc_2D_equals(backend):
    x1 = continuation_ldc_2D(backend)
    x2 = continuation_ldc_semi_2D(backend)

    dof1 = 3
    dof2 = 4

    assert utils.norm(x1[0:-1:dof1] - x2[0:-1:dof2]) < 1e-2
    assert utils.norm(x1[1:-1:dof1] - x2[1:-1:dof2]) < 1e-2
    assert utils.norm(x1[2:-1:dof1] - x2[3:-1:dof2]) < 1e-2

@pytest.mark.parametrize("backend", ["SciPy", "Epetra", "HYMLS", "PETSc"])
def test_continuation_ldc_2D_stretched(backend, nx=8):
    numpy.random.seed(1234)

    ny = nx

    parameters = {'Grid Stretching': True}
    interface = Interface(parameters, nx, ny, backend=backend)

    x = interface.discretization.get_coordinate_vector(0, 1, nx)
    y = interface.discretization.get_coordinate_vector(0, 1, ny)
    assert x[1] - x[0] < x[2] - x[1]
    assert y[1] - y[0] < y[2] - y[1]

    continuation = Continuation(interface)

    x0 = interface.vector()
    x0 = continuation.newton(x0)

    start = 0
    target = 2000
    ds = 100
    x = continuation.continuation(x0, 'Reynolds Number', start, target, ds)[0]

    assert utils.norm(x) > 0

def test_continuation_time_integration(nx=4):
    numpy.random.seed(1234)

    dof = 3
    ny = nx
    nz = 1

    parameters = {}
    interface = Interface(parameters, nx, ny, nz, dof=dof)
    continuation = Continuation(interface, newton_tolerance=1e-8)

    x0 = interface.vector()
    x0 = continuation.newton(x0)

    start = 0
    target = 2000
    ds = 100
    x = continuation.continuation(x0, 'Reynolds Number', start, target, ds)[0]

    # Start from a perturbed solution
    x2 = utils.create_state_mtx(x, nx, ny, nz, dof)
    x2[1:nx-1, 1:ny-1, :, 0] += 0.1 * numpy.random.random((nx - 2, ny - 2, nz))
    x2 = utils.create_state_vec(x2, nx, ny, nz, dof)

    assert numpy.linalg.norm(x[0:len(x):dof] - x2[0:len(x):dof]) > 1e-2
    assert numpy.linalg.norm(x[1:len(x):dof] - x2[1:len(x):dof]) < 1e-4

    time_integration = TimeIntegration(interface)
    x3 = time_integration.integration(x2, 1, 1)[0]

    assert numpy.linalg.norm(x[0:len(x):dof] - x3[0:len(x):dof]) > 1e-2
    assert numpy.linalg.norm(x[1:len(x):dof] - x3[1:len(x):dof]) > 1e-2

    x3 = time_integration.integration(x2, 100, 1000)[0]

    assert numpy.linalg.norm(x[0:len(x):dof] - x3[0:len(x):dof]) < 1e-4
    assert numpy.linalg.norm(x[1:len(x):dof] - x3[1:len(x):dof]) < 1e-4

    # Start from zero
    x2[:] = 0

    x3 = time_integration.integration(x2, 100, 1000)[0]

    assert numpy.linalg.norm(x[0:len(x):dof] - x3[0:len(x):dof]) < 1e-4
    assert numpy.linalg.norm(x[1:len(x):dof] - x3[1:len(x):dof]) < 1e-4

@pytest.mark.parametrize("backend", ["SciPy", "HYMLS"])
def test_continuation_dhc(backend, nx=8):
    try:
        from transiflow.interface import JaDa # noqa: F401
    except ImportError:
        pytest.skip('jadapy not found')

    numpy.random.seed(1234)

    ny = nx

    parameters = {'Problem Type': 'Differentially Heated Cavity',
                  'Rayleigh Number': 1,
                  'Prandtl Number': 1000,
                  'Reynolds Number': 1}

    interface = Interface(parameters, nx, ny, backend=backend)
    continuation = Continuation(interface, newton_tolerance=1e-9)

    x0 = interface.vector()

    start = 0
    target = 9e7
    ds = 1e4
    x, mu = continuation.continuation(x0, 'Rayleigh Number', start, target, ds, ds_max=1e7)

    parameters['Eigenvalue Solver'] = {}
    parameters['Eigenvalue Solver']['Number of Eigenvalues'] = 2

    target = 1e9
    ds = 5e6
    x, mu = continuation.continuation(x, 'Rayleigh Number', mu, target, ds, ds_max=5e6,
                                      detect_bifurcations=True)

    assert utils.norm(x) > 0
    assert mu > 9e7
    assert mu < target

@pytest.mark.parametrize("backend", ["SciPy", "HYMLS"])
def test_continuation_rayleigh_benard(backend, nx=8):
    try:
        from transiflow.interface import JaDa # noqa: F401
    except ImportError:
        pytest.skip('jadapy not found')

    numpy.random.seed(1234)

    ny = nx

    parameters = {'Problem Type': 'Rayleigh-Benard',
                  'Prandtl Number': 10,
                  'Biot Number': 1,
                  'X-max': 10,
                  'Bordered Solver': True}

    interface = Interface(parameters, nx, ny, backend=backend)
    continuation = Continuation(interface)

    x0 = interface.vector()
    x0 = continuation.newton(x0)

    start = 0
    target = 1800
    ds = 200
    x, mu = continuation.continuation(x0, 'Rayleigh Number', start, target, ds)

    parameters['Eigenvalue Solver'] = {}
    parameters['Eigenvalue Solver']['Arithmetic'] = 'real'
    parameters['Eigenvalue Solver']['Number of Eigenvalues'] = 2

    target = 5000
    ds = 50
    x2, mu2, dx, dmu = continuation.continuation(x, 'Rayleigh Number', mu, target, ds,
                                                 detect_bifurcations=True,
                                                 return_step=True)

    assert utils.norm(x2) > 0
    assert mu2 > mu
    assert mu2 < target

    x3, mu3 = continuation.continuation(x2, 'Rayleigh Number', mu2, target, ds,
                                        detect_bifurcations=True)

    assert utils.norm(x3) > 0
    assert mu3 > mu2
    assert mu3 < target

@pytest.mark.parametrize("backend", ["SciPy", "HYMLS"])
def test_continuation_rayleigh_benard_formulations_2D(backend, nx=8):
    try:
        from transiflow.interface import JaDa # noqa: F401
    except ImportError:
        pytest.skip('jadapy not found')

    numpy.random.seed(1234)

    dim = 2
    dof = 4
    ny = nx
    nz = 1

    parameters = {'Problem Type': 'Rayleigh-Benard',
                  'Prandtl Number': 10,
                  'Biot Number': 1,
                  'X-max': 10,
                  'Bordered Solver': True}

    interface = Interface(parameters, nx, ny, nz, dim, dof, backend=backend)
    continuation = Continuation(interface)

    x0 = interface.vector()
    x0 = continuation.newton(x0)

    start = 0
    target = 1800
    ds = 200
    x, mu = continuation.continuation(x0, 'Rayleigh Number', start, target, ds)

    parameters['Eigenvalue Solver'] = {}
    parameters['Eigenvalue Solver']['Arithmetic'] = 'real'
    parameters['Eigenvalue Solver']['Number of Eigenvalues'] = 2

    target = 5000
    ds = 50
    x2, mu2 = continuation.continuation(x, 'Rayleigh Number', mu, target, ds,
                                        detect_bifurcations=True)

    assert utils.norm(x2) > 0
    assert mu2 > mu
    assert mu2 < target

    parameters['Problem Type'] = 'Rayleigh-Benard Perturbation'

    # Subtract the motionless state
    y = interface.discretization.get_coordinate_vector(0, 1, ny)
    t = numpy.zeros((nx, ny, nz, dof))
    for j in range(ny):
        t[:, j, 0, dim+1] = 1 - (y[j] + y[j-1]) / 4
    t = utils.create_state_vec(t, nx, ny, nz, dof)
    x = x - interface.vector_from_array(t)

    x3, mu3 = continuation.continuation(x, 'Rayleigh Number', mu, target, ds,
                                        detect_bifurcations=True)

    # Test that the solution obtained from both formulations are the same
    x2 = interface.array_from_vector(x2)
    x3 = interface.array_from_vector(x3)

    assert utils.norm(x3[0:x.size:dof] - x2[0:x.size:dof]) < 1e-4
    assert utils.norm(x3[1:x.size:dof] - x2[1:x.size:dof]) < 1e-4
    assert utils.norm(x3[3:x.size:dof] - x2[3:x.size:dof] + t[3:x.size:dof]) < 1e-4
    assert mu3 > mu
    assert mu3 < target
    assert abs(mu3 - mu2) < 1

@pytest.mark.parametrize("backend", ["SciPy", "HYMLS"])
def test_continuation_rayleigh_benard_formulations(backend, nx=4):
    try:
        from transiflow.interface import JaDa # noqa: F401
    except ImportError:
        pytest.skip('jadapy not found')

    numpy.random.seed(1234)

    dim = 3
    dof = 5
    ny = nx
    nz = nx

    parameters = {'Problem Type': 'Rayleigh-Benard',
                  'Prandtl Number': 10,
                  'Biot Number': 1,
                  'X-max': 10,
                  'Y-max': 10,
                  'Bordered Solver': True}

    interface = Interface(parameters, nx, ny, nz, dim, dof, backend=backend)
    continuation = Continuation(interface)

    x0 = interface.vector()
    x0 = continuation.newton(x0)

    start = 0
    target = 1800
    ds = 200
    x, mu = continuation.continuation(x0, 'Rayleigh Number', start, target, ds)

    parameters['Eigenvalue Solver'] = {}
    parameters['Eigenvalue Solver']['Arithmetic'] = 'real'
    parameters['Eigenvalue Solver']['Number of Eigenvalues'] = 2

    target = 5000
    ds = 50
    x2, mu2 = continuation.continuation(x, 'Rayleigh Number', mu, target, ds,
                                        detect_bifurcations=True)

    assert utils.norm(x2) > 0
    assert mu2 > mu
    assert mu2 < target

    parameters['Problem Type'] = 'Rayleigh-Benard Perturbation'

    # Subtract the motionless state
    z = interface.discretization.get_coordinate_vector(0, 1, nz)
    t = numpy.zeros((nx, ny, nz, dof))
    for k in range(nz):
        t[:, :, k, dim+1] = 1 - (z[k] + z[k-1]) / 4
    t = utils.create_state_vec(t, nx, ny, nz, dof)
    x = x - interface.vector_from_array(t)

    x3, mu3 = continuation.continuation(x, 'Rayleigh Number', mu, target, ds,
                                        detect_bifurcations=True)

    # Test that the solution obtained from both formulations are the same
    x2 = interface.array_from_vector(x2)
    x3 = interface.array_from_vector(x3)

    assert utils.norm(x3[0:x.size:dof] - x2[0:x.size:dof]) < 1e-4
    assert utils.norm(x3[1:x.size:dof] - x2[1:x.size:dof]) < 1e-4
    assert utils.norm(x3[2:x.size:dof] - x2[2:x.size:dof]) < 1e-4
    assert utils.norm(x3[4:x.size:dof] - x2[4:x.size:dof] + t[4:x.size:dof]) < 1e-4
    assert mu3 > mu
    assert mu3 < target
    assert abs(mu3 - mu2) < 1

@pytest.mark.parametrize("backend", ["SciPy", "HYMLS"])
def test_continuation_double_gyre(backend, nx=8):
    try:
        from transiflow.interface import JaDa # noqa: F401
    except ImportError:
        pytest.skip('jadapy not found')

    numpy.random.seed(1234)

    ny = nx

    parameters = {'Problem Type': 'Double Gyre',
                  'Reynolds Number': 16,
                  'Rossby Parameter': 1000,
                  'Wind Stress Parameter': 0}

    interface = Interface(parameters, nx, ny, backend=backend)
    continuation = Continuation(interface)

    x0 = interface.vector()

    start = 0
    target = 1000
    ds = 200
    x, mu = continuation.continuation(x0, 'Wind Stress Parameter', start, target, ds)

    parameters['Eigenvalue Solver'] = {}
    parameters['Eigenvalue Solver']['Number of Eigenvalues'] = 2

    target = 100
    ds = 5
    x, mu = continuation.continuation(x, 'Reynolds Number', 16, target, ds,
                                      detect_bifurcations=True)

    assert utils.norm(x) > 0
    assert mu > 16
    assert mu < target

@pytest.mark.parametrize("backend", ["SciPy", "HYMLS"])
def test_continuation_amoc(backend, nx=16):
    try:
        from transiflow.interface import JaDa # noqa: F401
    except ImportError:
        pytest.skip('jadapy not found')

    numpy.random.seed(1234)

    ny = nx // 2

    parameters = {'Problem Type': 'AMOC',
                  'Rayleigh Number': 4e4,
                  'Prandtl Number': 2.25,
                  'X-max': 5}

    interface = Interface(parameters, nx, ny, backend=backend)
    continuation = Continuation(interface)

    x0 = interface.vector()

    target = 1
    ds = 0.1
    x, mu = continuation.continuation(x0, 'Temperature Forcing', 0, target, ds)

    parameters['Eigenvalue Solver'] = {}
    parameters['Eigenvalue Solver']['Number of Eigenvalues'] = 2

    target = 0.2
    ds = 0.01
    continuation = Continuation(interface, newton_tolerance=1e-6)
    x, mu = continuation.continuation(x, 'Freshwater Flux', 0, target,
                                      ds, ds_min=1e-6,
                                      detect_bifurcations=True)

    assert numpy.linalg.norm(x) > 0
    assert mu > 0.1
    assert mu < target

def test_continuation_2D_tc(nx=8):
    try:
        from transiflow.interface import JaDa # noqa: F401
    except ImportError:
        pytest.skip('jadapy not found')

    numpy.random.seed(1234)

    ny = 1
    nz = nx

    ri = 0.8
    ro = 1

    parameters = {'Problem Type': 'Taylor-Couette',
                  'Reynolds Number': 1,
                  'R-min': ri,
                  'R-max': ro,
                  'Z-max': 1,
                  'Z-periodic': True,
                  'Inner Angular Velocity': 1 / ri / (ro - ri),
                  'Outer Angular Velocity': 0}

    interface = Interface(parameters, nx, ny, nz)
    continuation = Continuation(interface)

    x0 = interface.vector()

    start = 0
    target = 80
    ds = 30
    x, mu = continuation.continuation(x0, 'Reynolds Number', start, target, ds)

    parameters['Bordered Solver'] = True
    parameters['Eigenvalue Solver'] = {}
    parameters['Eigenvalue Solver']['Number of Eigenvalues'] = 2
    parameters['Eigenvalue Solver']['Arithmetic'] = 'real'

    target = 100
    ds = 1
    continuation = Continuation(interface, newton_tolerance=1e-12)
    x, mu = continuation.continuation(x, 'Reynolds Number', mu, target, ds,
                                      detect_bifurcations=True)

    assert numpy.linalg.norm(x) > 0
    assert mu > 0
    assert mu < target



from transiflow.utils import norm

class TimeIntegration:
    '''Time integration using the theta method.

    Parameters
    ----------
    interface
        Interface object that implements the following functions:
        ``rhs(x)``, ``jacobian(x)``, ``mass_matrix()`` and
        ``solve(jac, rhs)``.
    newton_tolerance : scalar, optional
        Tolerance used in the Newton corrector to determine
        convergence.
    maximum_newton_iterations : int, optional
        Maximum number of Newton iterations.
    residual_check: str, optional
        Method for checking the residual in the Newton method
        (default: 'F').

        - ``F``: Use the norm of the ``rhs(x)`` function.
        - ``dx``: Use the norm of the step ``dx``.
    verbose : bool, optional
        Give extra information about convergence. Since we have to
        compute this information, this may be slower.

    '''

    def __init__(self, interface, theta=1.0,
                 newton_tolerance=1e-10, maximum_newton_iterations=10,
                 residual_check='F', verbose=False):
        self.interface = interface
        self.verbose = verbose
        self.theta = theta

        self.maximum_newton_iterations = maximum_newton_iterations
        self.newton_tolerance = newton_tolerance
        self.residual_check = residual_check

    def _newton(self, x0, dt):
        x = x0
        b0 = self.interface.rhs(x0)
        mass = self.interface.mass_matrix()

        for k in range(self.maximum_newton_iterations):
            # M * u_n + dt * theta * F(u_(n+1)) + dt * (1 - theta) * F(u_n) - M * u_(n+1) = 0
            fval = mass @ (x0 - x) + dt * self.theta * self.interface.rhs(x) + dt * (1 - self.theta) * b0
            fval /= self.theta * dt

            if self.residual_check == 'F' or self.verbose:
                fnorm = norm(fval)

            if self.residual_check == 'F' and fnorm < self.newton_tolerance:
                print('Newton converged i %d iterations with ||F||=%e' % (k, fnorm), flush=True)
                break

            # J - 1 / (theta * dt) * M
            jac = self.interface.jacobian(x) - mass / (self.theta * dt)
            dx = self.interface.solve(jac, -fval)

            x = x + dx

            if self.residual_check != 'F' or self.verbose:
                dxnorm = norm(dx)

            if self.residual_check != 'F' and dxnorm < self.newton_tolerance:
                print('Newton converged in %d iterations with ||dx||=%e' % (k, dxnorm), flush=True)
                break

            if self.verbose:
                print('Newton status at iteration %d: ||F||=%e, ||dx||=%e' % (k, fnorm, dxnorm), flush=True)

        return x

    def integration(self, x0, dt, tmax, callback=None):
        '''
        Perform the time integration.

        Parameters
        ----------
        x0 : array_like
            Initial solution.
        dt : scalar
            Time step.
        tmax : scalar
            End time.
        callback : function, optional
            User-supplied function to call after each continuation
            step. It is called as ``callback(interface, x, t)``.

        Returns
        -------
        x : array_like
            Solution at the end time
        t : scalar
            End time

        '''

        x = x0
        t = 0

        if callback:
            callback(self.interface, x, t)

        while t < tmax:
            x = self._newton(x, dt)
            t += dt

            print("t = %f" % t, flush=True)

            if callback:
                callback(self.interface, x, t)

        return x, t



import numpy
import petsc4py
from petsc4py import PETSc

from transiflow.interface import ParallelBaseInterface

petsc4py.init()


class Vector(PETSc.Vec):
    """Distributed ghosted PETSc Vector with some extra methods added
    to it for convenience."""

    @staticmethod
    def from_array(m, x, ghosts=(), comm=PETSc.COMM_WORLD):
        size = len(m.indices)
        vec = Vector().createGhost(ghosts, (size, None), comm=comm)
        vec.setValues(m.indices, x)
        vec.assemble()
        vec.ghostUpdate(addv=PETSc.InsertMode.INSERT_VALUES, mode=PETSc.ScatterMode.FORWARD)
        return vec

    size = property(PETSc.Vec.getSize)

    def gather(self):
        sct, vector_global = PETSc.Scatter().toZero(self)
        sct.scatter(self, vector_global)
        return vector_global

    def all_gather(self):
        sct, vector_global = PETSc.Scatter().toAll(self)
        sct.scatter(self, vector_global)
        return vector_global

    @property
    def shape(self):
        return (self.size,)


class Interface(ParallelBaseInterface):
    """This class defines an interface to the PETSc backend for the
    discretization. This backend can be used for testing
    low-resolution parallel runs as well as a base for a class that
    implements a better linear solver. It should not be used for any
    meaningful simulations.

    The PETSc backend partitions the domain into Cartesian subdomains. The
    subdomains will be distributed over multiple processors if MPI is
    used to run the application.

    PETSc parallel vectors (and matrices) are distributed such that
    each process owns a contiguous range of indices (rows) without
    holes. The natural ordering is therefore mapped to PETSc's
    ordering.

    See :mod:`.Discretization` for the descriptions of the constructor
    arguments.

    Parameters
    ----------
    parameters : dict
        Key-value pairs that can be used to modify parameters in the
        discretization as well as the linear solver and eigenvalue
        solver.

    """

    def __init__(self, parameters, nx, ny, nz=1, dim=None, dof=None,
                 boundary_conditions=None, comm=PETSc.COMM_WORLD):
        super().__init__(comm, parameters, nx, ny, nz, dim, dof, boundary_conditions)

        self.size_global = nx * ny * nz * self.dof

        self.map_natural = self.create_map()
        self.assembly_map = self.create_map(True)

        ghosts_natural = list(
            set(self.assembly_map.indices).difference(self.map_natural.indices)
        )

        indices_natural = Vector().createWithArray(self.map_natural.indices)
        sct, indices_natural_global = PETSc.Scatter().toAll(indices_natural)
        sct.scatter(indices_natural, indices_natural_global)
        self.index_natural_global_permut = numpy.argsort(indices_natural_global)

        self.ghosts = self._ghosts_petsc(ghosts_natural, indices_natural_global)

        self.index_ordering = PETSc.AO().createMapping(self.map_natural.indices)
        self.map = self._create_map_petsc()

        self.index_natural_local_permut = numpy.argsort(
            numpy.r_[self.map_natural.indices, ghosts_natural]
        )

        self.index_ordering_assembly = PETSc.AO().createMapping(
            numpy.r_[self.map_natural.indices, ghosts_natural].astype(PETSc.IntType),
            numpy.r_[self.map.indices, self.ghosts].astype(PETSc.IntType),
        )

        self.jac = None

    def vector(self):
        return Vector().createGhost(self.ghosts, (len(self.map.indices), None), comm=self.comm)

    def vector_from_array(self, array):
        return Vector.from_array(self.map, array[self.map_natural.indices], ghosts=self.ghosts)

    def array_from_vector(self, vector):
        return Vector.all_gather(vector).array[self.index_natural_global_permut]

    def _ghosts_petsc(self, ghosts_natural, indices_natural_global):
        """Get global indices of ghosts in PETSc ordering."""
        ghosts_petsc = []
        for ghost in ghosts_natural:
            ghosts_petsc.append(numpy.where(indices_natural_global.array == ghost)[0][0])

        return ghosts_petsc

    def _create_map_petsc(self):
        """Create a PETSc index map, using PETSc's ordering (contiguous row ownership),
        from the naturally ordered indices."""
        return PETSc.LGMap().create(
            self.index_ordering.app2petsc(self.map_natural.indices), bsize=None, comm=self.comm
        )

    def create_map(self, overlapping=False):
        """Create a PETSc map on which the local discretization domain is defined.
        The overlapping part is only used for computing the discretization.

        :meta private:

        """
        local_elements = ParallelBaseInterface.create_map(self, overlapping)
        return PETSc.LGMap().create(local_elements, bsize=None, comm=self.comm)

    def rhs(self, state):
        r"""Compute the right-hand side of the DAE. That is the
        right-hand side $F(u, p)$ in

        .. math:: M(p) \frac{\d u}{\d t} = F(u, p)

        The RHS is computed on the overlapping subdomain map and then
        distributed using the non-overlapping map that is used by the
        linear solver.

        Parameters
        ----------
        state : array_like
            State $u$ at which to evaluate $F(u, p)$.

        Returns
        -------
        rhs : array_like
            The value of $F(u, p)$.

        """
        state.ghostUpdate(addv=PETSc.InsertMode.INSERT_VALUES,
                          mode=PETSc.ScatterMode.FORWARD)
        ind_local_nat = self.index_natural_local_permut

        with state.localForm() as lf:
            local_rhs = self.discretization.rhs(lf.array[ind_local_nat])

        rhs_vec = state.duplicate()

        with rhs_vec.localForm() as lf:
            lf.array[ind_local_nat] = local_rhs

        return rhs_vec

    def jacobian(self, state):
        r"""Compute the Jacobian matrix $J(u, p)$ of the right-hand
        side of the DAE. That is the Jacobian matrix of $F(u, p)$ in

        .. math:: M(p) \frac{\d u}{\d t} = F(u, p)

        The Jacobian matrix is computed on the overlapping subdomain
        map and then distributed using the non-overlapping map that is
        used by the linear solver.

        Parameters
        ----------
        state : array_like
            State $u$ at which to evaluate $J(u, p)$.

        Returns
        -------
        jac : PETSc.Mat
            The matrix $J(u, p)$ in a CSR matrix format that allows
            for distributing the overlap.

        """
        ind_local_nat = self.index_natural_local_permut

        state.ghostUpdate(addv=PETSc.InsertMode.INSERT_VALUES,
                          mode=PETSc.ScatterMode.FORWARD)

        with state.localForm() as lf:
            local_jac = self.discretization.jacobian(lf.array[ind_local_nat])

        if self.jac is None:
            self.jac = PETSc.Mat().create(comm=self.comm)
            self.jac.setType("aij")
            N = state.size
            self.jac.setSizes((N, N))
            self.jac.setLGMap(self.map)
            self.jac.setUp()
        else:
            self.jac.zeroEntries()

        self.jac.assemblyBegin()
        for i_loc in range(len(local_jac.begA) - 1):

            if self.is_ghost(i_loc):
                continue

            col_idx_nat = self.assembly_map.indices[
                local_jac.jcoA[local_jac.begA[i_loc]: local_jac.begA[i_loc + 1]]
            ]
            values = local_jac.coA[local_jac.begA[i_loc]: local_jac.begA[i_loc + 1]]

            i = self.index_ordering.app2petsc(self.assembly_map.indices[i_loc])
            col_idx = self.index_ordering_assembly.app2petsc(col_idx_nat)
            self.jac.setValues(i, col_idx, values)

        self.jac.assemblyEnd()

        return self.jac

    def solve(self, jac, rhs):
        ksp = PETSc.KSP().create()
        ksp.setType("preonly")
        pc = ksp.getPC()
        pc.setType("lu")
        pc.setFactorSolverType("mumps")

        ksp.setFromOptions()

        dof = self.dim
        self._set_dirichlet_bc(jac, rhs, dof, 0.0)

        ksp.setOperators(jac)
        x = self.vector()

        ksp.solve(rhs, x)

        x.ghostUpdate(addv=PETSc.InsertMode.INSERT_VALUES, mode=PETSc.ScatterMode.FORWARD)
        return x

    def _set_dirichlet_bc(self, A, b, i, value):
        """Set Dirichlet BC to specified value at index i.
        Single values or sequences of indices and values are accepted.
        """
        A.zeroRows(i)
        b.setValues(i, value)
        b.assemble()
        b.ghostUpdate(addv=PETSc.InsertMode.INSERT_VALUES, mode=PETSc.ScatterMode.FORWARD)



import numpy

from .utils import create_state_vec

class BoundaryConditions:
    '''Boundary conditions for the Discretization class. Every method
    expects and atom, modifies it accordingly and also modifies the
    forcing that is stored internally. This class is only used from
    within the Discretization class.

    Parameters
    ----------
    nx : int
        Grid size in the x direction.
    ny : int
        Grid size in the y direction.
    nz : int
        Grid size in the z direction. This is set to 1 for
        2-dimensional problems.
    dim : int
        Physical dimension of the problem. In case this is set to 2, w
        is not referenced in the state vector.
    dof : int
        Degrees of freedom for this problem. This should be set to dim
        plus 1 for each of pressure, temperature and salinity, if they
        are required for your problem. For example a 3D differentially
        heated cavity has dof = 3 + 1 + 1 = 5.
    x : array_like, optional
        Coordinate vector in the x direction.
    y : array_like, optional
        Coordinate vector in the y direction.
    z : array_like, optional
        Coordinate vector in the z direction.

    '''

    def __init__(self, nx, ny, nz, dim, dof, x, y, z):

        self.nx = nx
        self.ny = ny
        self.nz = nz
        self.dim = dim
        self.dof = dof

        self.x = x
        self.y = y
        self.z = z

        self.frc = numpy.zeros((self.nx, self.ny, self.nz, self.dof))

    def get_forcing(self):
        '''Return the forcing that is required to be added the the RHS.'''
        return create_state_vec(self.frc, self.nx, self.ny, self.nz, self.dof)

    def no_slip_east(self, atom):
        '''Apply a no slip boundary condition at the east boundary. At this
        boundary $u_i = 0$, $(v_i + v_{i+1}) / 2 = 0$ and similar for
        $w$. So we replace $v_{i+1}$ by $-v_i$, $w_{i+1}$ by $-w_i$,
        set $u_i$ to zero and set all references to cells outside the
        domain to zero as well.

        '''
        atom[self.nx-1, :, :, :, :, 1, :, :] -= atom[self.nx-1, :, :, :, :, 2, :, :]
        atom[self.nx-1, :, :, :, 0, 1, :, :] = 0
        atom[self.nx-1, :, :, 0, :, :, :, :] = 0
        atom[self.nx-1, :, :, :, :, 2, :, :] = 0
        atom[self.nx-1, :, :, 0, 0, 1, 1, 1] = -1
        # TODO: Do we need this?
        atom[self.nx-2, :, :, 0, 0, 2, :, :] = 0

        self.frc[self.nx-1, :, :, 0] = 0

    def no_slip_west(self, atom):
        '''Apply a no slip boundary condition at the west boundary. At this
        boundary $u_{i-1} = 0$, $(v_{i-1} + v_i) / 2 = 0$ and similar
        for $w$. So we simply replace $v_{i-1}$ by $-v_i$, $w_{i-1}$
        by $-w_i$ and set all references to cells outside the domain
        to zero.

        '''
        atom[0, :, :, :, 0, 0, :, :] = 0
        atom[0, :, :, :, :, 1, :, :] -= atom[0, :, :, :, :, 0, :, :]
        atom[0, :, :, :, :, 0, :, :] = 0

    def no_slip_north(self, atom):
        '''Apply a no slip boundary condition at the north boundary. At this
        boundary $v_j = 0$, $(u_j + u_{j+1}) / 2 = 0$ and similar for
        $w$. So we replace $u_{j+1}$ by $-u_j$, $w_{j+1}$ by $-w_j$,
        set $v_j$ to zero and set all references to cells outside the
        domain to zero as well.

        '''
        atom[:, self.ny-1, :, :, :, :, 1, :] -= atom[:, self.ny-1, :, :, :, :, 2, :]
        atom[:, self.ny-1, :, :, 1, :, 1, :] = 0
        atom[:, self.ny-1, :, 1, :, :, :, :] = 0
        atom[:, self.ny-1, :, :, :, :, 2, :] = 0
        atom[:, self.ny-1, :, 1, 1, 1, 1, 1] = -1
        # TODO: Do we need this?
        atom[:, self.ny-2, :, 1, 1, :, 2, :] = 0

        self.frc[:, self.ny-1, :, 1] = 0

    def no_slip_south(self, atom):
        '''Apply a no slip boundary condition at the south boundary. At this
        boundary $v_{j-1} = 0$, $(u_{j-1} + u_j) / 2 = 0$ and similar
        for $w$. So we simply replace $u_{j-1}$ by $-u_j$, $w_{j-1}$
        by $-w_j$ and set all references to cells outside the domain
        to zero.

        '''
        atom[:, 0, :, :, 1, :, 0, :] = 0
        atom[:, 0, :, :, :, :, 1, :] -= atom[:, 0, :, :, :, :, 0, :]
        atom[:, 0, :, :, :, :, 0, :] = 0

    def no_slip_top(self, atom):
        '''Apply a no slip boundary condition at the top boundary. At this
        boundary $w_k = 0$, $(u_k + u_{k+1}) / 2 = 0$ and similar for
        $v$. So we replace $u_{k+1}$ by $-u_k$, $v_{k+1}$ by $-v_k$,
        set $w_k$ to zero and set all references to cells outside the
        domain to zero as well.

        '''
        atom[:, :, self.nz-1, :, :, :, :, 1] -= atom[:, :, self.nz-1, :, :, :, :, 2]
        atom[:, :, self.nz-1, :, 2, :, :, 1] = 0
        atom[:, :, self.nz-1, 2, :, :, :, :] = 0
        atom[:, :, self.nz-1, :, :, :, :, 2] = 0
        atom[:, :, self.nz-1, 2, 2, 1, 1, 1] = -1
        # TODO: Do we need this?
        atom[:, :, self.nz-2, 2, 2, :, :, 2] = 0

        self.frc[:, :, self.nz-1, 1] = 0

    def no_slip_bottom(self, atom):
        '''Apply a no slip boundary condition at the bottom boundary. At this
        boundary $w_{k-1} = 0$, $(u_{k-1} + u_k) / 2 = 0$ and similar
        for $v$. So we simply replace $u_{k-1}$ by $-u_k$, $v_{k-1}$
        by $-v_k$, and set all references to cells outside the domain
        to zero.

        '''
        atom[:, :, 0, :, 2, :, :, 0] = 0
        atom[:, :, 0, :, :, :, :, 1] -= atom[:, :, 0, :, :, :, :, 0]
        atom[:, :, 0, :, :, :, :, 0] = 0

    def free_slip_east(self, atom):
        '''Apply a free slip boundary condition at the east boundary. At this
        boundary $u_i = 0$, $(v_{i+1} - v_i) / h = 0$ and similar for
        $w$. So we replace $v_{i+1}$ by $v_i$, $w_{i+1}$ by $w_i$, set
        $u_i$ to zero and set all references to cells outside the
        domain to zero as well.

        '''
        atom[self.nx-1, :, :, :, :, 1, :, :] += atom[self.nx-1, :, :, :, :, 2, :, :]
        atom[self.nx-1, :, :, :, 0, 1, :, :] = 0
        atom[self.nx-1, :, :, 0, :, :, :, :] = 0
        atom[self.nx-1, :, :, :, :, 2, :, :] = 0
        atom[self.nx-1, :, :, 0, 0, 1, 1, 1] = -1
        # TODO: Do we need this?
        atom[self.nx-2, :, :, 0, 0, 2, :, :] = 0

        self.frc[self.nx-1, :, :, 0] = 0

    def free_slip_west(self, atom):
        '''Apply a free slip boundary condition at the west boundary. At this
        boundary $u_{i-1} = 0$, $(v_i - v_{i-1}) / h = 0$ and similar
        for $w$. So we simply replace $v_{i-1}$ by $v_i$, $w_{i-1}$ by
        $w_i$ and set all references to cells outside the domain to
        zero as well.

        '''
        atom[0, :, :, :, 0, 0, :, :] = 0
        atom[0, :, :, :, :, 1, :, :] += atom[0, :, :, :, :, 0, :, :]
        atom[0, :, :, :, :, 0, :, :] = 0

    def free_slip_north(self, atom):
        '''Apply a free slip boundary condition at the north boundary. At this
        boundary $v_j = 0$, $(u_{j+1} - u_j) / h = 0$ and similar for
        $w$. So we replace $u_{j+1}$ by $u_j$, $w_{j+1}$ by $w_j$, set
        $v_j$ to zero and set all references to cells outside the
        domain to zero as well.

        '''
        atom[:, self.ny-1, :, :, :, :, 1, :] += atom[:, self.ny-1, :, :, :, :, 2, :]
        atom[:, self.ny-1, :, :, 1, :, 1, :] = 0
        atom[:, self.ny-1, :, 1, :, :, :, :] = 0
        atom[:, self.ny-1, :, :, :, :, 2, :] = 0
        atom[:, self.ny-1, :, 1, 1, 1, 1, 1] = -1
        # TODO: Do we need this?
        atom[:, self.ny-2, :, 1, 1, :, 2, :] = 0

        self.frc[:, self.ny-1, :, 1] = 0

    def free_slip_south(self, atom):
        '''Apply a free slip boundary condition at the south boundary. At this
        boundary $v_{j-1} = 0$, $(u_j - u_{j-1}) / h = 0$ and similar
        for $w$. So we simply replace $u_{j-1}$ by $u_j$, $w_{j-1}$ by
        $w_j$ and set all references to cells outside the domain to
        zero as well.

        '''
        atom[:, 0, :, :, 1, :, 0, :] = 0
        atom[:, 0, :, :, :, :, 1, :] += atom[:, 0, :, :, :, :, 0, :]
        atom[:, 0, :, :, :, :, 0, :] = 0

    def free_slip_top(self, atom):
        '''Apply a free slip boundary condition at the top boundary. At this
        boundary $w_k = 0$, $(u_{k+1} - u_k) / h = 0$ and similar for
        $v$. So we replace $u_{k+1}$ by $u_k$, $v_{k+1}$ by $v_k$, set
        $w_k$ to zero and set all references to cells outside the
        domain to zero as well.

        '''
        atom[:, :, self.nz-1, :, :, :, :, 1] += atom[:, :, self.nz-1, :, :, :, :, 2]
        atom[:, :, self.nz-1, :, 2, :, :, 1] = 0
        atom[:, :, self.nz-1, 2, :, :, :, :] = 0
        atom[:, :, self.nz-1, :, :, :, :, 2] = 0
        atom[:, :, self.nz-1, 2, 2, 1, 1, 1] = -1
        # TODO: Do we need this?
        atom[:, :, self.nz-2, 2, 2, :, :, 2] = 0

        self.frc[:, :, self.nz-1, 1] = 0

    def free_slip_bottom(self, atom):
        '''Apply a free slip boundary condition at the south boundary. At this
        boundary $w_{k-1} = 0$, $(u_k - u_{k-1}) / h = 0$ and similar
        for $v$. So we simply replace $u_{k-1}$ by $u_k$, $v_{k-1}$ by
        $v_k$ and set all references to cells outside the domain to
        zero as well.

        '''
        atom[:, :, 0, :, 2, :, :, 0] = 0
        atom[:, :, 0, :, :, :, :, 1] += atom[:, :, 0, :, :, :, :, 0]
        atom[:, :, 0, :, :, :, :, 0] = 0

    def moving_lid_east(self, atom, velocity):
        '''Apply a moving lid at the east boundary with velocity V. At this
        boundary $u_i = 0$, $(v_i + v_{i+1}) / 2 = V$ and no slip for
        $w$. So we replace $v_{i+1}$ by $2V - v_i$ and then apply the no
        slip condition.

        '''
        self.frc += self._constant_forcing_east(atom, 1, 2 * velocity, -1)
        self.no_slip_east(atom)

    def moving_lid_west(self, atom, velocity):
        '''Apply a moving lid at the west boundary with velocity V. At this
        boundary $u_i = 0$, $(v_i + v_{i-1}) / 2 = V$ and no slip for
        $w$. So we replace $v_{i-1}$ by $2V - v_i$ and then apply the no
        slip condition.

        '''
        self.frc += self._constant_forcing_west(atom, 1, 2 * velocity, -1)
        self.no_slip_west(atom)

    def moving_lid_north(self, atom, velocity):
        '''Apply a moving lid at the north boundary with velocity U. At this
        boundary $v_j = 0$, $(u_j + u_{j+1}) / 2 = U$ and no slip for
        $w$. So we replace $u_{j+1}$ by $2U - u_j$ and then apply the no
        slip condition.

        '''
        self.frc += self._constant_forcing_north(atom, 0, 2 * velocity, -1)
        self.no_slip_north(atom)

    def moving_lid_south(self, atom, velocity):
        '''Apply a moving lid at the south boundary with velocity U. At this
        boundary $v_j = 0$, $(u_j + u_{j-1}) / 2 = U$ and no slip for
        $w$. So we replace $u_{j-1}$ by $2U - u_j$ and then apply the no
        slip condition.

        '''
        self.frc += self._constant_forcing_south(atom, 0, 2 * velocity, -1)
        self.no_slip_south(atom)

    def moving_lid_top(self, atom, velocity):
        '''Apply a moving lid at the top boundary with velocity U. At this
        boundary $w_k = 0$, $(u_k + u_{k+1}) / 2 = U$ and no slip for
        $v$. So we replace $u_{k+1}$ by $2U - u_k$ and then apply the no
        slip condition.

        '''
        self.frc += self._constant_forcing_top(atom, 0, 2 * velocity, -1)
        self.no_slip_top(atom)

    def moving_lid_bottom(self, atom, velocity):
        '''Apply a moving lid at the bottom boundary with velocity U. At this
        boundary $w_k = 0$, $(u_k + u_{k-1}) / 2 = U$ and no slip for
        $v$. So we replace $u_{k-1}$ by $2U - u_k$ and then apply the no
        slip condition.

        '''
        self.frc += self._constant_forcing_bottom(atom, 0, 2 * velocity, -1)
        self.no_slip_bottom(atom)

# TODO: These methods are untested with nonzero heat flux and stretched grids

    def temperature_east(self, atom, temperature):
        '''Apply a fixed temperature $T_b$ at the east boundary. We have $(T_i
        + T_{i+1}) / 2 = T_b$ so we replace $T_{i+1}$ by $2T_b - T_i$.

        '''
        self.frc += self._constant_forcing_east(atom, self.dim+1, 2 * temperature, -1)

    def temperature_west(self, atom, temperature):
        '''Apply a fixed temperature $T_b$ at the west boundary. We have $(T_i
        + T_{i-1}) / 2 = T_b$ so we replace $T_{i-1}$ by $2T_b - T_i$.

        '''
        self.frc += self._constant_forcing_west(atom, self.dim+1, 2 * temperature, -1)

    def temperature_north(self, atom, temperature):
        '''Apply a fixed temperature $T_b$ at the north boundary. We have $(T_j
        + T_{j+1}) / 2 = T_b$ so we replace $T_{j+1}$ by $2T_b - T_j$.

        '''
        self.frc += self._constant_forcing_north(atom, self.dim+1, 2 * temperature, -1)

    def temperature_south(self, atom, temperature):
        '''Apply a fixed temperature $T_b$ at the south boundary. We have $(T_j
        + T_{j-1}) / 2 = T_b$ so we replace $T_{j-1}$ by $2T_b - T_j$.

        '''
        self.frc += self._constant_forcing_south(atom, self.dim+1, 2 * temperature, -1)

    def temperature_top(self, atom, temperature):
        '''Apply a fixed temperature $T_b$ at the top boundary. We have $(T_k
        + T_{k+1}) / 2 = T_b$ so we replace $T_{k+1}$ by $2T_b - T_k$.

        '''
        self.frc += self._constant_forcing_top(atom, self.dim+1, 2 * temperature, -1)

    def temperature_bottom(self, atom, temperature):
        '''Apply a fixed temperature $T_b$ at the bottom boundary. We have $(T_k
        + T_{k-1}) / 2 = T_b$ so we replace $T_{k-1}$ by $2T_b - T_k$.

        '''
        self.frc += self._constant_forcing_bottom(atom, self.dim+1, 2 * temperature, -1)

    def heat_flux_east(self, atom, heat_flux, biot=0.0):
        r'''Apply a heat flux $Q$ at the east boundary. We have $(T_{i+1} -
        T_i) / h - \\mathrm{Bi} (T_{i+1} + T_i) / 2 = Q$ so we get
        $T_{i+1} = T_i (1 - h \\mathrm{Bi} / 2) / (1 + h \\mathrm{Bi} /
        2) + h Q / (1 + h \\mathrm{Bi} / 2)$. By default, the Biot
        number is assumed to be zero.

        '''
        h = (self.x[self.nx] - self.x[self.nx-2]) / 2

        forcing_constant = h * heat_flux / (1 + h * biot / 2)
        atom_constant = (1 - h * biot / 2) / (1 + h * biot / 2)
        self.frc += self._constant_forcing_east(atom, self.dim+1, forcing_constant, atom_constant)

    def heat_flux_west(self, atom, heat_flux, biot=0.0):
        r'''Apply a heat flux $Q$ at the west boundary. We have $(T_i -
        T_{i-1}) / h - \\mathrm{Bi} (T_{i-1} + T_i) / 2 = Q$ so we get
        $T_{i-1} = T_i (1 + h \\mathrm{Bi} / 2) / (1 - h \\mathrm{Bi}
        / 2) + h Q / (1 - h \\mathrm{Bi} / 2)$. By default, the Biot
        number is assumed to be zero.

        '''
        h = (self.x[0] - self.x[-2]) / 2

        forcing_constant = -h * heat_flux / (1 - h * biot / 2)
        atom_constant = (1 + h * biot / 2) / (1 - h * biot / 2)
        self.frc += self._constant_forcing_west(atom, self.dim+1, forcing_constant, atom_constant)

    def heat_flux_north(self, atom, heat_flux, biot=0.0):
        r'''Apply a heat flux $Q$ at the north boundary. We have $(T_{j+1} -
        T_j) / h - \\mathrm{Bi} (T_{j+1} + T_j) / 2 = Q$ so we get
        $T_{j+1} = T_j (1 - h \\mathrm{Bi} / 2) / (1 + h \\mathrm{Bi}
        / 2) + h Q / (1 + h \\mathrm{Bi} / 2)$. By default, the Biot
        number is assumed to be zero.

        '''
        h = (self.y[self.ny] - self.y[self.ny-2]) / 2

        forcing_constant = h * heat_flux / (1 + h * biot / 2)
        atom_constant = (1 - h * biot / 2) / (1 + h * biot / 2)
        self.frc += self._constant_forcing_north(atom, self.dim+1, forcing_constant, atom_constant)

    def heat_flux_south(self, atom, heat_flux, biot=0.0):
        r'''Apply a heat flux $Q$ at the south boundary. We have $(T_j -
        T_{j-1}) / h - \\mathrm{Bi} (T_{j-1} + T_i) / 2 = Q$ so we get
        $T_{j-1} = T_j (1 + h \\mathrm{Bi} / 2) / (1 - h \\mathrm{Bi}
        / 2) + h Q / (1 - h \\mathrm{Bi} / 2)$. By default, the Biot
        number is assumed to be zero.

        '''
        h = (self.y[0] - self.y[-2]) / 2

        forcing_constant = -h * heat_flux / (1 - h * biot / 2)
        atom_constant = (1 + h * biot / 2) / (1 - h * biot / 2)
        self.frc += self._constant_forcing_south(atom, self.dim+1, forcing_constant, atom_constant)

    def heat_flux_top(self, atom, heat_flux, biot=0.0):
        r'''Apply a heat flux $Q$ at the top boundary. We have $(T_{k+1} -
        T_k) / h - \\mathrm{Bi} (T_{k+1} + T_k) / 2 = Q$ so we get
        $T_{k+1} = T_k (1 - h \\mathrm{Bi} / 2) / (1 + h \\mathrm{Bi}
        / 2) + h Q / (1 + h \\mathrm{Bi} / 2)$. By default, the Biot
        number is assumed to be zero.

        '''
        h = (self.z[self.nz] - self.z[self.nz-2]) / 2

        forcing_constant = h * heat_flux / (1 + h * biot / 2)
        atom_constant = (1 - h * biot / 2) / (1 + h * biot / 2)
        self.frc += self._constant_forcing_top(atom, self.dim+1, forcing_constant, atom_constant)

    def heat_flux_bottom(self, atom, heat_flux, biot=0.0):
        r'''Apply a heat flux $Q$ at the bottom boundary. We have $(T_k -
        T_{k-1}) / h - \\mathrm{Bi} (T_{k-1} + T_i) / 2 = Q$ so we get
        $T_{k-1} = T_k (1 + h \\mathrm{Bi} / 2) / (1 - h \\mathrm{Bi}
        / 2) + h Q / (1 - h \\mathrm{Bi} / 2)$. By default, the Biot
        number is assumed to be zero.

        '''
        h = (self.z[0] - self.z[-2]) / 2

        forcing_constant = -h * heat_flux / (1 - h * biot / 2)
        atom_constant = (1 + h * biot / 2) / (1 - h * biot / 2)
        self.frc += self._constant_forcing_bottom(atom, self.dim+1, forcing_constant, atom_constant)

    def salinity_flux_east(self, atom, salinity_flux):
        '''Apply a salinity flux $Q$ at the east boundary. We have $(S_{i+1} -
        S_i) / h = Q$ so we replace $S_{i+1}$ by $S_i + hQ$.

        '''
        h = (self.x[self.nx] - self.x[self.nx-2]) / 2
        self.frc += self._constant_forcing_east(atom, self.dim+2, h * salinity_flux, 1)

    def salinity_flux_west(self, atom, salinity_flux):
        '''Apply a salinity flux $Q$ at the west boundary. We have $(S_i -
        S_{i-1}) / h = Q$ so we replace $S_{i-1}$ by $S_i - hQ$.

        '''
        h = (self.x[0] - self.x[-2]) / 2
        self.frc += self._constant_forcing_west(atom, self.dim+2, -h * salinity_flux, 1)

    def salinity_flux_north(self, atom, salinity_flux):
        '''Apply a salinity flux $Q$ at the north boundary. We have $(S_{j+1} -
        S_j) / h = Q$ so we replace $S_{j+1}$ by $S_j + hQ$.

        '''
        h = (self.y[self.ny] - self.y[self.ny-2]) / 2
        self.frc += self._constant_forcing_north(atom, self.dim+2, h * salinity_flux, 1)

    def salinity_flux_south(self, atom, salinity_flux):
        '''Apply a salinity flux $Q$ at the south boundary. We have $(S_j -
        S_{j-1}) / h = Q$ so we replace $S_{j-1}$ by $S_j - hQ$.

        '''
        h = (self.y[0] - self.y[-2]) / 2
        self.frc += self._constant_forcing_south(atom, self.dim+2, -h * salinity_flux, 1)

    def salinity_flux_top(self, atom, salinity_flux):
        '''Apply a salinity flux $Q$ at the top boundary. We have $(S_{k+1} -
        S_k) / h = Q$ so we replace $S_{k+1}$ by $S_k + hQ$.

        '''
        h = (self.z[self.nz] - self.z[self.nz-2]) / 2
        self.frc += self._constant_forcing_top(atom, self.dim+2, h * salinity_flux, 1)

    def salinity_flux_bottom(self, atom, salinity_flux):
        '''Apply a salinity flux $Q$ at the bottom boundary. We have $(S_k -
        S_{k-1}) / h = Q$ so we replace $S_{k-1}$ by $S_k - hQ$.

        '''
        h = (self.z[0] - self.z[-2]) / 2
        self.frc += self._constant_forcing_bottom(atom, self.dim+2, -h * salinity_flux, 1)

    def _constant_forcing(self, atom, nx, ny, var, value):
        value = numpy.ones((nx + 2, ny + 2)) * value

        frc = numpy.zeros((nx, ny, self.dof))
        for j, i, y, x in numpy.ndindex(ny, nx, 3, 3):
            frc[i, j, var] += atom[i, j, var, x, y] * value[i + x - 1, j + y - 1]

        return frc

    def _constant_forcing_east(self, atom, var, forcing_constant, atom_constant):
        frc = numpy.zeros((self.nx, self.ny, self.nz, self.dof))
        frc[self.nx-1, :, :, :] = self._constant_forcing(
            atom[self.nx-1, :, :, :, var, 2, :, :], self.ny, self.nz,
            var, forcing_constant)

        atom[self.nx-1, :, :, :, var, 1, :, :] += atom_constant * atom[self.nx-1, :, :, :, var, 2, :, :]
        atom[self.nx-1, :, :, :, var, 2, :, :] = 0

        return frc

    def _constant_forcing_west(self, atom, var, forcing_constant, atom_constant):
        frc = numpy.zeros((self.nx, self.ny, self.nz, self.dof))
        frc[0, :, :, :] = self._constant_forcing(
            atom[0, :, :, :, var, 0, :, :], self.ny, self.nz,
            var, forcing_constant)

        atom[0, :, :, :, var, 1, :, :] += atom_constant * atom[0, :, :, :, var, 0, :, :]
        atom[0, :, :, :, var, 0, :, :] = 0

        return frc

    def _constant_forcing_north(self, atom, var, forcing_constant, atom_constant):
        frc = numpy.zeros((self.nx, self.ny, self.nz, self.dof))
        frc[:, self.ny-1, :, :] = self._constant_forcing(
            atom[:, self.ny-1, :, :, var, :, 2, :], self.nx, self.nz,
            var, forcing_constant)

        atom[:, self.ny-1, :, :, var, :, 1, :] += atom_constant * atom[:, self.ny-1, :, :, var, :, 2, :]
        atom[:, self.ny-1, :, :, var, :, 2, :] = 0

        return frc

    def _constant_forcing_south(self, atom, var, forcing_constant, atom_constant):
        frc = numpy.zeros((self.nx, self.ny, self.nz, self.dof))
        frc[:, 0, :, :] = self._constant_forcing(
            atom[:, 0, :, :, var, :, 0, :], self.nx, self.nz,
            var, forcing_constant)

        atom[:, 0, :, :, var, :, 1, :] += atom_constant * atom[:, 0, :, :, var, :, 0, :]
        atom[:, 0, :, :, var, :, 0, :] = 0

        return frc

    def _constant_forcing_top(self, atom, var, forcing_constant, atom_constant):
        frc = numpy.zeros((self.nx, self.ny, self.nz, self.dof))
        frc[:, :, self.nz-1, :] = self._constant_forcing(
            atom[:, :, self.nz-1, :, var, :, :, 2], self.nx, self.ny,
            var, forcing_constant)

        atom[:, :, self.nz-1, :, var, :, :, 1] += atom_constant * atom[:, :, self.nz-1, :, var, :, :, 2]
        atom[:, :, self.nz-1, :, var, :, :, 2] = 0

        return frc

    def _constant_forcing_bottom(self, atom, var, forcing_constant, atom_constant):
        frc = numpy.zeros((self.nx, self.ny, self.nz, self.dof))
        frc[:, :, 0, :] = self._constant_forcing(
            atom[:, :, 0, :, var, :, :, 0], self.nx, self.ny,
            var, forcing_constant)

        atom[:, :, 0, :, var, :, :, 1] += atom_constant * atom[:, :, 0, :, var, :, :, 0]
        atom[:, :, 0, :, var, :, :, 0] = 0

        return frc



import numpy

from transiflow.interface.SciPy import Interface


def test_solve(nx=4):
    dim = 3
    dof = 4
    ny = nx
    nz = nx

    parameters = {}
    interface = Interface(parameters, nx, ny, nz, dim, dof)

    n = dof * nx * ny * nz

    x = numpy.random.random(n)

    x0 = numpy.zeros(n)
    A = interface.jacobian(x0)

    b = A @ x

    y = interface.solve(A, b)

    assert numpy.linalg.norm(y) > 0

    y[dim::dof] += x[dim]
    assert numpy.linalg.norm(y) > 0
    assert numpy.linalg.norm(y - x) < 1e-11

def test_iterative_solve(nx=4):
    dim = 3
    dof = 4
    ny = nx
    nz = nx

    parameters = {'Use Iterative Solver': True,
                  'Iterative Solver': {'Convergence Tolerance': 1e-8}}
    interface = Interface(parameters, nx, ny, nz, dim, dof)

    n = dof * nx * ny * nz

    x = numpy.random.random(n)

    x0 = numpy.zeros(n)
    A = interface.jacobian(x0)

    b = A @ x

    y = interface.solve(A, b)

    print(y)
    print(x)

    assert numpy.linalg.norm(b) > 0
    assert numpy.linalg.norm(y) > 0

    tol = parameters['Iterative Solver']['Convergence Tolerance'] * numpy.linalg.norm(b) * 10

    y[dim::dof] += x[dim]
    assert numpy.linalg.norm(y) > 0
    assert numpy.linalg.norm(y - x) > tol / 10 # The pressure is inaccurate
    assert numpy.linalg.norm(y[0::dof] - x[0::dof]) < tol
    assert numpy.linalg.norm(y[1::dof] - x[1::dof]) < tol
    assert numpy.linalg.norm(y[2::dof] - x[2::dof]) < tol

def test_bordered_matrix(nx=4):
    dim = 3
    dof = 4
    ny = nx
    nz = nx

    parameters = {}
    interface = Interface(parameters, nx, ny, nz, dim, dof)
    interface.border_scaling = 1

    n = dof * nx * ny * nz
    n2 = 3

    x = numpy.random.random(n)
    x2 = numpy.random.random(n2)
    x3 = numpy.append(x, x2)

    V = numpy.random.random((n, n2))
    W = numpy.random.random((n, n2))
    C = numpy.random.random((n2, n2))

    x0 = numpy.zeros(n)
    A = interface.jacobian(x0)

    b = A @ x + V @ x2
    b2 = W.T @ x + C @ x2
    b3 = numpy.append(b, b2)

    B = interface.compute_bordered_matrix(A, V, W, C)

    y = B @ x3

    assert numpy.linalg.norm(y) > 0
    assert numpy.linalg.norm(y - b3) < 1e-11



import os

import numpy
import pytest

from transiflow import Continuation, Discretization, utils


def read_matrix(fname, ao):
    from petsc4py import PETSc

    dirname = os.path.dirname(__file__)
    rows = []
    cols = []
    vals = []
    with open(os.path.join(dirname, fname), "r") as f:
        for i in f.readlines():
            r, c, v = [j.strip() for j in i.strip().split(" ") if j]
            r = int(r) - 1
            c = int(c) - 1
            v = float(v)
            rows.append(r)
            cols.append(c)
            vals.append(v)

    size = max(rows) + 1
    A = PETSc.Mat().createAIJ((size, size))
    A.setUp()

    for r, c, v in zip(ao.app2petsc(rows), ao.app2petsc(cols), vals):
        A.setValue(r, c, v)
    A.assemble()

    return A


def read_vector(fname, m, ao):
    from petsc4py import PETSc

    from transiflow.interface import PETSc as PETScInterface

    values = []
    dirname = os.path.dirname(__file__)
    with open(os.path.join(dirname, fname), "r") as f:
        for v in f.readlines():
            values.append(float(v.strip()))

    indices_nat = ao.petsc2app(m.indices)
    vec = PETScInterface.Vector.from_array(
        m, numpy.array(values, dtype=PETSc.ScalarType)[indices_nat]
    )
    return vec


def extract_sorted_row(A, i):
    if i in range(*A.getOwnershipRange()):
        indices, values = A.getRow(i)
        idx = sorted(range(len(indices)), key=lambda i: indices[i])
        return [indices[i] for i in idx], [values[i] for i in idx]
    else:
        return [], []


def extract_sorted_local_row(A, i, ao):
    indices = A.jcoA[A.begA[i]: A.begA[i + 1]]
    values = A.coA[A.begA[i]: A.begA[i + 1]]
    indices = ao.app2petsc(indices.astype(numpy.int32))
    idx = sorted(range(len(indices)), key=lambda i: indices[i])
    return [indices[i] for i in idx], [values[i] for i in idx]


def test_ldc():
    try:
        from transiflow.interface import PETSc as PETScInterface
    except ImportError:
        pytest.skip("PETSc not found")

    nx = 4
    ny = nx
    nz = nx
    dof = 4
    parameters = {"Reynolds Number": 100}
    n = nx * ny * nz * dof

    state = numpy.zeros(n)
    for i in range(n):
        state[i] = i + 1

    interface = PETScInterface.Interface(parameters, nx, ny, nz, dof=dof)
    state = interface.vector_from_array(state)

    for i in interface.map.indices:
        i_nat = interface.index_ordering.petsc2app(i)
        print(i, i_nat, state[i] - 1)
        assert i_nat == state[i] - 1

    A = interface.jacobian(state)
    rhs = interface.rhs(state)

    B = read_matrix("data/ldc_%sx%sx%s.txt" % (nx, ny, nz), interface.index_ordering_assembly)
    rhs_B = read_vector(
        "data/ldc_rhs_%sx%sx%s.txt" % (nx, ny, nz), interface.map, interface.index_ordering
    )

    for i in range(n):
        if i not in interface.map.indices:
            continue

        indices_A, values_A = extract_sorted_row(A, i)
        indices_B, values_B = extract_sorted_row(B, i)

        print("Expected:")
        print(indices_B)
        print(values_B)

        print("Got:")
        print(indices_A)
        print(values_A)

        assert len(indices_A) == len(indices_B) > 0

        for j in range(len(indices_A)):
            assert indices_A[j] == indices_B[j]
            assert values_A[j] == pytest.approx(values_B[j])

    for i in range(n):
        if i not in interface.map.indices:
            continue

        print(i, rhs[i], rhs_B[i])
        assert rhs_B[i] == pytest.approx(rhs[i])


def test_ldc_stretched_file():
    try:
        from petsc4py import PETSc

        from transiflow.interface import PETSc as PETScInterface
    except ImportError:
        pytest.skip("PETSc not found")

    nx = 4
    ny = nx
    nz = nx
    dof = 4
    parameters = {"Reynolds Number": 100, "Grid Stretching": True}
    n = nx * ny * nz * dof

    state = numpy.zeros(n, dtype=PETSc.ScalarType)
    for i in range(n):
        state[i] = i + 1

    interface = PETScInterface.Interface(parameters, nx, ny, nz, dof=dof)
    state = interface.vector_from_array(state)

    A = interface.jacobian(state)
    rhs = interface.rhs(state)

    B = read_matrix(
        "data/ldc_stretched_%sx%sx%s.txt" % (nx, ny, nz), interface.index_ordering_assembly
    )
    rhs_B = read_vector(
        "data/ldc_stretched_rhs_%sx%sx%s.txt" % (nx, ny, nz), interface.map, interface.index_ordering
    )

    for i in range(n):
        if i not in interface.map.indices:
            continue

        indices_A, values_A = extract_sorted_row(A, i)
        indices_B, values_B = extract_sorted_row(B, i)

        print("Expected:")
        print(indices_B)
        print(values_B)

        print("Got:")
        print(indices_A)
        print(values_A)

        assert len(indices_A) == len(indices_B) > 0

        for j in range(len(indices_A)):
            assert indices_A[j] == indices_B[j]
            assert values_A[j] == pytest.approx(values_B[j])

    for i in range(n):
        if i not in interface.map.indices:
            continue

        print(i, rhs[i], rhs_B[i])
        assert rhs_B[i] == pytest.approx(rhs[i])


def test_ldc_stretched(nx=4):
    try:
        from petsc4py import PETSc

        from transiflow.interface import PETSc as PETScInterface
    except ImportError:
        pytest.skip("PETSc not found")

    ny = nx
    nz = nx
    dof = 4
    parameters = {"Reynolds Number": 100, "Grid Stretching": True}
    n = nx * ny * nz * dof

    state = numpy.zeros(n, dtype=PETSc.ScalarType)
    for i in range(n):
        state[i] = i + 1

    discretization = Discretization(parameters, nx, ny, nz, dof=dof)
    B = discretization.jacobian(state)
    rhs_B = discretization.rhs(state)

    interface = PETScInterface.Interface(parameters, nx, ny, nz, dof=dof)

    state_numpy = state.copy()

    state = interface.vector_from_array(state)

    for i in interface.map.indices:
        i_nat = interface.index_ordering.petsc2app(i)
        print(i, i_nat, state[i] - 1)
        assert i_nat == state[i] - 1

    assert numpy.allclose(state.array, state_numpy[interface.map_natural.indices])

    for i in range(n):
        if i not in interface.map.indices:
            continue

        i_nat = interface.index_ordering.petsc2app(i)
        print(i, state[i], i_nat, state_numpy[i_nat])
        assert state_numpy[i_nat] == pytest.approx(state[i])

    # check ghosts mapping and assembly index ordering
    with state.localForm() as lf:
        for i, ghost in enumerate(interface.ghosts):
            ghost_nat = interface.index_ordering_assembly.petsc2app(ghost)
            print(i, ghost, lf.array[state.local_size + i])
            print(i, ghost_nat, state_numpy[ghost_nat])
            assert state_numpy[ghost_nat] == lf.array[state.local_size + i]

    A = interface.jacobian(state)
    rhs = interface.rhs(state)

    for i in range(n):
        if i not in interface.map.indices:
            continue

        i_nat = interface.index_ordering.petsc2app(i)
        print(i, rhs[i], i_nat, rhs_B[i_nat])
        assert rhs_B[i_nat] == pytest.approx(rhs[i])

    for i in range(n):
        if i not in interface.map.indices:
            continue

        indices_A, values_A = extract_sorted_row(A, i)

        i_nat = interface.index_ordering_assembly.petsc2app(i)

        # the assembly & solve map index orderings map i to the same PETSc index
        j_nat = interface.index_ordering.petsc2app(i)
        print(i, i_nat, j_nat)
        assert i_nat == j_nat

        indices_B, values_B = extract_sorted_local_row(
            B, i_nat, interface.index_ordering_assembly
        )

        print("Expected:")
        print(indices_B)
        print(values_B)

        print("Got:")
        print(indices_A)
        print(values_A)

        assert len(indices_A) == len(indices_B) > 0

        for j in range(len(indices_A)):
            assert indices_A[j] == indices_B[j]
            assert values_A[j] == pytest.approx(values_B[j])


def test_ldc8_stretched():
    test_ldc_stretched(8)


def test_norm():
    try:
        from transiflow.interface import PETSc as PETScInterface
    except ImportError:
        pytest.skip("PETSc not found")

    nx = 4
    ny = nx
    nz = nx
    dof = 4
    n = nx * ny * nz * dof
    parameters = {}

    state = numpy.zeros(n)
    for i in range(n):
        state[i] = i + 1

    interface = PETScInterface.Interface(parameters, nx, ny, nz, dof=dof)

    state_dist = PETScInterface.Vector.from_array(
        interface.map, state[interface.map_natural.indices]
    )
    state_dist2 = interface.vector_from_array(state)

    assert utils.norm(state) == utils.norm(state_dist)
    assert utils.norm(state) == utils.norm(state_dist2)


def test_PETSc(nx=4):
    try:
        from transiflow.interface import PETSc as PETScInterface
    except ImportError:
        pytest.skip("PETSc not found")

    numpy.random.seed(1234)

    ny = nx
    nz = nx
    parameters = {"Reynolds Number": 0, "Verbose": True}

    interface = PETScInterface.Interface(parameters, nx, ny, nz)

    continuation = Continuation(interface)

    x0 = interface.vector()
    x0 = continuation.newton(x0)

    start = 0
    target = 2000
    ds = 100
    x = continuation.continuation(x0, "Reynolds Number", start, target, ds)[0]

    assert utils.norm(x) > 0

def test_vector(nx=8):
    try:
        from transiflow.interface import PETSc as PETScInterface
    except ImportError:
        pytest.skip("PETSc not found")

    ny = nx
    nz = nx
    dim = 3
    dof = 4
    parameters = {"Reynolds Number": 100, "Grid Stretching": True}
    n = nx * ny * nz * dof

    state = numpy.zeros(n)
    for i in range(n):
        state[i] = i + 1

    print(state)

    interface = PETScInterface.Interface(parameters, nx, ny, nz, dim, dof)
    state_vec = interface.vector_from_array(state)
    state_gathered = interface.array_from_vector(state_vec)

    print(state_gathered)

    assert state == pytest.approx(state_gathered)



import numpy

from transiflow.interface import BaseInterface


def ind2sub(nx, ny, nz, idx, dof=1):
    rem = idx
    var = rem % dof
    rem = rem // dof
    i = rem % nx
    rem = rem // nx
    j = rem % ny
    rem = rem // ny
    k = rem % nz
    return (i, j, k, var)


def sub2ind(nx, ny, nz, dof, i, j, k, var):
    return ((k * ny + j) * nx + i) * dof + var


def get_local_coordinate_vector(x, nx_offset, nx_local):
    x = numpy.roll(x, 2)
    x = x[nx_offset:nx_offset + nx_local + 3]
    return numpy.roll(x, -2)


class ParallelBaseInterface(BaseInterface):
    '''This is the base class for all parallel interfaces. It
    implements methods for creating a Cartesian partitioning of the
    domain that can be used for computing local discretizations with
    the required overlap. The subdomains will be distributed over
    multiple processors if MPI is used to run the application.

    '''
    def __init__(self, comm, parameters, nx, ny, nz=1, dim=None, dof=None,
                 boundary_conditions=None):
        super().__init__(parameters, nx, ny, nz, dim, dof,
                         boundary_conditions=boundary_conditions)

        self.nx_global = self.nx
        self.ny_global = self.ny
        self.nz_global = self.nz

        self.comm = comm

        self.partition_domain()

        self.discretization.x = get_local_coordinate_vector(self.discretization.x,
                                                            self.nx_offset, self.nx_local)
        self.discretization.y = get_local_coordinate_vector(self.discretization.y,
                                                            self.ny_offset, self.ny_local)
        self.discretization.z = get_local_coordinate_vector(self.discretization.z,
                                                            self.nz_offset, self.nz_local)

        self.discretization.nx = self.nx_local
        self.discretization.ny = self.ny_local
        self.discretization.nz = self.nz_local

    def get_comm_size(self):
        '''Get the total number of processors.'''
        try:
            return self.comm.size
        except AttributeError:
            return self.comm.NumProc()

    def get_comm_rank(self):
        '''Get the rank of the current processor.'''
        try:
            return self.comm.rank
        except AttributeError:
            return self.comm.MyPID()

    def save_json(self, name, obj):
        if self.get_comm_rank() == 0:
            BaseInterface.save_json(self, name, obj)

    def save_state(self, name, x):
        x = self.array_from_vector(x)

        if self.get_comm_rank() == 0:
            BaseInterface.save_state(self, name, x)

        self.comm.Barrier()

    def load_state(self, name):
        x = BaseInterface.load_state(self, name)
        return self.vector_from_array(x)

    def partition_domain(self):
        '''Partition the domain into Cartesian subdomains for computing the
        discretization.

        :meta private:

        '''

        rmin = 1e100

        self.npx = 1
        self.npy = 1
        self.npz = 1

        nparts = self.get_comm_size()
        pid = self.get_comm_rank()

        found = False

        # check all possibilities of splitting the map
        for t1 in range(1, nparts + 1):
            for t2 in range(1, nparts // t1 + 1):
                t3 = nparts // (t1 * t2)
                if t1 * t2 * t3 == nparts:
                    nx_loc = self.nx_global // t1
                    ny_loc = self.ny_global // t2
                    nz_loc = self.nz_global // t3

                    if (nx_loc * t1 != self.nx_global or ny_loc * t2 != self.ny_global
                            or nz_loc * t3 != self.nz_global):
                        continue

                    r1 = abs(self.nx_global / t1 - self.ny_global / t2)
                    r2 = abs(self.nx_global / t1 - self.nz_global / t3)
                    r3 = abs(self.ny_global / t2 - self.nz_global / t3)
                    r = r1 + r2 + r3

                    if r < rmin:
                        rmin = r
                        self.npx = t1
                        self.npy = t2
                        self.npz = t3
                        found = True

        if not found:
            raise Exception("Could not split %dx%dx%d domain in %d parts." %
                            (self.nx_global, self.ny_global, self.nz_global, nparts))

        self.pidx, self.pidy, self.pidz, _ = ind2sub(self.npx, self.npy, self.npz, pid)

        # Compute the local domain size and offset.
        self.nx_local = self.nx_global // self.npx
        self.ny_local = self.ny_global // self.npy
        self.nz_local = self.nz_global // self.npz

        self.nx_offset = self.nx_local * self.pidx
        self.ny_offset = self.ny_local * self.pidy
        self.nz_offset = self.nz_local * self.pidz

        # Add ghost nodes to factor out boundary conditions in the interior.
        if self.pidx > 0:
            self.nx_local += 2
            self.nx_offset -= 2
        if self.pidx < self.npx - 1:
            self.nx_local += 2
        if self.pidy > 0:
            self.ny_local += 2
            self.ny_offset -= 2
        if self.pidy < self.npy - 1:
            self.ny_local += 2
        if self.pidz > 0:
            self.nz_local += 2
            self.nz_offset -= 2
        if self.pidz < self.npz - 1:
            self.nz_local += 2

    def is_ghost(self, i, j=None, k=None):
        '''If a node is a ghost node that is used only for computing the
        discretization and is located outside of an interior boundary.

        :meta private:

        '''

        if j is None:
            i, j, k, _ = ind2sub(self.nx_local, self.ny_local, self.nz_local, i, self.dof)

        ghost = False
        if self.pidx > 0 and i < 2:
            ghost = True
        elif self.pidx < self.npx - 1 and i >= self.nx_local - 2:
            ghost = True
        elif self.pidy > 0 and j < 2:
            ghost = True
        elif self.pidy < self.npy - 1 and j >= self.ny_local - 2:
            ghost = True
        elif self.pidz > 0 and k < 2:
            ghost = True
        elif self.pidz < self.npz - 1 and k >= self.nz_local - 2:
            ghost = True

        return ghost

    def create_map(self, overlapping=False):
        '''Create a map on which the local discretization domain is defined.
        The overlapping part is only used for computing the discretization.

        :meta private:

        '''

        local_elements = [0] * self.nx_local * self.ny_local * self.nz_local * self.dof

        pos = 0
        for k in range(self.nz_local):
            for j in range(self.ny_local):
                for i in range(self.nx_local):
                    if not overlapping and self.is_ghost(i, j, k):
                        continue
                    for var in range(self.dof):
                        local_elements[pos] = sub2ind(
                            self.nx_global,
                            self.ny_global,
                            self.nz_global,
                            self.dof,
                            i + self.nx_offset,
                            j + self.ny_offset,
                            k + self.nz_offset,
                            var,
                        )
                        pos += 1

        return local_elements[0:pos]



import numpy
import os
import pytest

from transiflow import utils
from transiflow.interface import create


def Interface(*args, backend="SciPy", **kwargs):
    try:
        return create(*args, backend=backend, **kwargs)
    except ImportError:
        pytest.skip(backend + " not found")


@pytest.mark.parametrize("backend", ["SciPy", "Epetra", "HYMLS", "PETSc"])
def test_solve(backend, nx=4):
    ny = nx
    nz = nx

    parameters = {}
    interface = Interface(parameters, nx, ny, nz, backend="SciPy")

    x0 = interface.vector()
    n = x0.size

    A = interface.jacobian(x0)

    x = numpy.zeros(n)
    for i in range(n):
        x[i] = i + 1

    pressure_node = 3
    x[pressure_node] = 0

    b = A @ x

    interface = Interface(parameters, nx, ny, nz, backend=backend)

    x0 = interface.vector()
    A = interface.jacobian(x0)
    b = interface.vector_from_array(b)

    y = interface.solve(A, b)
    y = interface.array_from_vector(y)

    pressure = y[pressure_node] - x[pressure_node]

    assert utils.norm(y) > 0
    assert (utils.norm(y - x) - pressure * nx * ny * nz) / utils.norm(b) < 1e-7

def test_bordered_matrix(nx=4):
    ny = nx
    nz = nx

    parameters = {}
    interface = Interface(parameters, nx, ny, nz)
    interface.border_scaling = 1

    x0 = interface.vector()
    n = x0.size
    n2 = 3

    x = numpy.random.random(n)
    x2 = numpy.random.random(n2)
    x3 = numpy.append(x, x2)

    V = numpy.random.random((n, n2))
    W = numpy.random.random((n, n2))
    C = numpy.random.random((n2, n2))

    A = interface.jacobian(x0)

    b = A @ x + V @ x2
    b2 = W.T @ x + C @ x2
    b3 = numpy.append(b, b2)

    B = interface.compute_bordered_matrix(A, V, W, C)

    y = B @ x3

    assert numpy.linalg.norm(y) > 0
    assert numpy.linalg.norm(y - b3) < 1e-11


@pytest.mark.parametrize("backend", ["SciPy", "Epetra", "HYMLS", "PETSc"])
def test_save_load(backend, nx=4):
    ny = nx
    nz = nx

    try:
        from mpi4py import MPI
        if MPI.COMM_WORLD.Get_size() > 1 and backend == 'SciPy':
            pytest.skip('SciPy shouldn\'t be used in parallel')
    except ImportError:
        pass

    parameters = {'Eigenvalue Solver': {'Target': 1 + 3j}}
    interface = Interface(parameters, nx, ny, nz, backend=backend)

    x = interface.vector()
    n = x.size

    x = interface.vector_from_array(numpy.random.random(n))
    interface.save_state('x-test', x)

    assert os.path.isfile('x-test.npy')
    assert os.path.isfile('x-test.params')

    parameters['Eigenvalue Solver']['Target'] = 33

    x2 = interface.load_state('x-test')
    assert parameters['Eigenvalue Solver']['Target'] == 1 + 3j
    assert utils.norm(x - x2) < 1e-14


@pytest.mark.parametrize("backend", ["SciPy", "Epetra", "HYMLS", "PETSc"])
def test_custom_bc(backend, nx=4):
    def boundaries(boundary_conditions, atom):
        boundary_conditions.heat_flux_east(atom, 0)
        boundary_conditions.heat_flux_west(atom, 0)
        boundary_conditions.no_slip_east(atom)
        boundary_conditions.no_slip_west(atom)

        boundary_conditions.heat_flux_north(atom, 0)
        boundary_conditions.heat_flux_south(atom, 0)
        boundary_conditions.no_slip_north(atom)
        boundary_conditions.no_slip_south(atom)

        Bi = 0
        boundary_conditions.heat_flux_top(atom, 0, Bi)
        boundary_conditions.temperature_bottom(atom, 0)
        boundary_conditions.free_slip_top(atom)
        boundary_conditions.no_slip_bottom(atom)

        return boundary_conditions.get_forcing()

    ny = nx
    nz = nx
    dim = 3
    dof = 5

    parameters = {'Rayleigh Number': 100, 'Prandtl Number': 100}
    interface = Interface(parameters, nx, ny, nz, dim, dof, backend=backend,
                          boundary_conditions=boundaries)

    x = interface.vector()
    interface.rhs(x)



import time
import numpy
import warnings

from jadapy import NumPyInterface
from jadapy.orthogonalization import orthogonalize

from transiflow.interface.SciPy import gmres


def _get_scalars(alpha, beta):
    try:
        if len(alpha.shape) == 2:
            alpha = alpha[0, 0]
            beta = beta[0, 0]
        elif len(alpha.shape) == 1:
            alpha = alpha[0]
            beta = beta[0]
    except AttributeError:
        pass

    return alpha, beta

class Op:
    def __init__(self, mat):
        self.mat = mat
        self.dtype = mat.data.dtype
        self.shape = mat.shape

    def matvec(self, x):
        return self.mat * x

    def __matmul__(self, x):
        return self.mat * x

class CachedMatrix:
    def __init__(self, matrix, alpha, beta):
        self.matrix = matrix
        self.alpha = alpha
        self.beta = beta
        self.last_used = time.time()

    def same_shifts(self, alpha, beta):
        eps = 1e-10
        eps2 = 1e-1

        if abs(alpha) < eps:
            if abs(alpha - self.alpha) < eps2 and abs(beta - self.beta) < eps2:
                return True
        elif abs(beta) > eps and abs(self.beta) > eps:
            if abs(alpha / beta - self.alpha / self.beta) / abs(alpha / beta) < eps2:
                return True

        return False

    def same_shapes(self, matrix):
        return matrix is None or self.matrix.shape == matrix.shape

    def get_matrix(self):
        self.last_used = time.time()
        return self.matrix

class MatrixCache:
    def __init__(self, jac_op, mass_op):
        self.jac_op = jac_op
        self.mass_op = mass_op

        self.matrices = []
        self.max_matrices = 1

    def get_shifted_matrix(self, alpha, beta, shifted_matrix=None):
        alpha, beta = _get_scalars(alpha, beta)

        if shifted_matrix is None and alpha == 0.0 and beta == 1.0:
            return self.jac_op.mat

        # Cache previous preconditioners
        for i, cached_matrix in enumerate(self.matrices):
            if cached_matrix.same_shifts(alpha, beta) and cached_matrix.same_shapes(shifted_matrix):
                return cached_matrix.get_matrix()

        # Remove the cached preconditioner that was not used last
        if len(self.matrices) >= self.max_matrices:
            if len(self.matrices) > 1 and self.matrices[0].last_used > self.matrices[1].last_used:
                self.matrices.pop(1)
            elif len(self.matrices) > 0:
                self.matrices.pop(0)

        if shifted_matrix is None:
            shifted_matrix = beta * self.jac_op.mat - alpha * self.mass_op.mat

        if self.max_matrices > 0:
            self.matrices.append(CachedMatrix(shifted_matrix, alpha, beta))

        return shifted_matrix

class PrecOp(object):
    def __init__(self, op, prec_op):
        self.op = op
        self.prec_op = prec_op

        self.dtype = self.op.dtype
        self.shape = self.op.shape

    def matvec(self, x):
        return self.op.proj(self.prec_op(x, self.op.alpha, self.op.beta))

class Interface(NumPyInterface.NumPyInterface):
    def __init__(self, interface, jac_op, mass_op, *args, **kwargs):
        super().__init__(*args)
        self.interface = interface
        self.jac_op = jac_op

        self.preconditioned_solve = kwargs.get('preconditioned_solve', False)
        self.shifted = kwargs.get('shifted', False)

        self._matrix_cache = MatrixCache(jac_op, mass_op)

    def solve(self, op, x, tol, maxit):
        if op.dtype.char != op.dtype.char.upper():
            # Real case
            if abs(op.alpha.real) < abs(op.alpha.imag):
                op.alpha = op.alpha.imag
            else:
                op.alpha = op.alpha.real
            op.beta = op.beta.real

        prec_op = None
        if self.preconditioned_solve:
            if self.shifted:
                prec_op = PrecOp(op, self.shifted_prec)
            else:
                prec_op = PrecOp(op, self.prec)

        out = x.copy()
        for i in range(x.shape[1]):
            y, info, iterations = gmres(op, x[:, i], maxit, tol, prec=prec_op)
            out[:, i] = op.proj(y)

            if info < 0:
                raise Exception('GMRES returned ' + str(info))
            elif info > 0 and maxit > 1:
                warnings.warn('GMRES did not converge in ' + str(iterations) + ' iterations')
        return out

    def prec(self, x, *args):
        return self.interface.solve(self.jac_op.mat, x)

    def shifted_prec(self, x, alpha, beta):
        shifted_matrix = self._matrix_cache.get_shifted_matrix(alpha, beta)
        return self.interface.solve(shifted_matrix, x)

class BorderedPrecOp(object):
    def __init__(self, interface, prec):
        self.interface = interface
        self.prec = prec

        self.dtype = prec.dtype
        self.shape = prec.shape

    def matvec(self, x):
        return self.interface.solve(self.prec, x)

class BorderedInterface(NumPyInterface.NumPyInterface):
    def __init__(self, interface, jac_op, mass_op, *args, **kwargs):
        super().__init__(*args)
        self.interface = interface
        self.jac_op = jac_op
        self.mass_op = mass_op

        self._matrix_cache = MatrixCache(jac_op, mass_op)

    def solve(self, op, x, tol, maxit):
        alpha = op.alpha
        beta = op.beta

        if op.dtype.char != op.dtype.char.upper():
            # Real case
            if abs(op.alpha.real) < abs(op.alpha.imag):
                alpha = op.alpha.imag
            else:
                alpha = op.alpha.real
            beta = op.beta.real

        alpha, beta = _get_scalars(alpha, beta)

        shifted_matrix = beta * self.jac_op.mat - alpha * self.mass_op.mat
        shifted_bordered_matrix = self.interface.compute_bordered_matrix(shifted_matrix, op.Z, op.Q)
        shifted_bordered_op = Op(shifted_bordered_matrix)

        prec = self._matrix_cache.get_shifted_matrix(alpha, beta, shifted_bordered_matrix)
        prec_op = BorderedPrecOp(self.interface, prec)

        out = x.copy()
        for i in range(x.shape[1]):
            x2 = numpy.append(x[:, i], numpy.zeros(op.Q.shape[1], x.dtype))
            y, info, iterations = gmres(shifted_bordered_op, x2, maxit, tol, prec=prec_op)
            out[:, i] = op.proj(y[:-op.Q.shape[1]])

            if info < 0:
                raise Exception('GMRES returned ' + str(info))
            elif info > 0 and maxit > 1:
                warnings.warn('GMRES did not converge in ' + str(iterations) + ' iterations')

        orthogonalize(op.Q, out)

        return out

    def prec(self, x, *args):
        return self.interface.solve(self.jac_op.mat, x)



import numpy
import pytest

from numpy.testing import assert_allclose

import matplotlib.pyplot as plt

from transiflow import Continuation

# Import common fixtures
from tests.jada_fixtures import * # noqa: F401, F403

@pytest.fixture(autouse=True, scope='module')
def import_test():
    try:
        from transiflow.interface import JaDa # noqa: F401
    except ImportError:
        pytest.skip('jadapy not found')

    try:
        from transiflow.interface import HYMLS as HYMLSInterface # noqa: F401
    except ImportError:
        pytest.skip('HYMLS not found')

@pytest.fixture(scope='module')
def interface(nx):
    from transiflow.interface import HYMLS as HYMLSInterface
    from PyTrilinos import Teuchos

    dim = 2
    ny = nx
    nz = 1

    parameters = Teuchos.ParameterList()
    parameters.set('Reynolds Number', 0)
    parameters.set('Bordered Solver', True)

    interface = HYMLSInterface.Interface(parameters, nx, ny, nz, dim)

    return interface

@pytest.fixture(scope='module')
def x(interface):
    from transiflow.interface import HYMLS as HYMLSInterface

    continuation = Continuation(interface)

    x0 = HYMLSInterface.Vector(interface.map)
    x0 = continuation.newton(x0)

    start = 0
    target = 2000
    ds = 100
    return continuation.continuation(x0, 'Reynolds Number', start, target, ds)[0]

class Operator:
    def __init__(self, A):
        self.A = A

    def matvec(self, x):
        return self.A @ x

    def proj(self, x):
        return x

def test_solve(interface, x, tol):
    from transiflow.interface import HYMLS as HYMLSInterface
    from transiflow.interface import JaDaHYMLS

    from jadapy import EpetraInterface
    from jadapy.utils import norm

    from PyTrilinos import Epetra

    interface.teuchos_parameters.sublist('Preconditioner').set('Number of Levels', 0)
    interface.initialize()

    x_sol = HYMLSInterface.Vector(interface.solve_map)
    x_sol.Import(x, interface.solve_importer, Epetra.Insert)

    # Create a test vector to remove the nonzero first pressure
    t = HYMLSInterface.Vector(x_sol)
    t.PutScalar(0.0)
    for i in range(t.MyLength()):
        if t.Map().GID(i) != interface.dim:
            t[i] = 1.0

    jac_op = EpetraInterface.CrsMatrix(interface.jacobian(x))
    jada_interface = JaDaHYMLS.Interface(interface, preconditioned_solve=True)

    op = Operator(jac_op)
    b = op.matvec(x_sol)

    x = jada_interface.solve(op, b, tol, maxit=1)

    r = op.matvec(x) - b

    # Remove the nonzero first pressure
    r.Multiply(1.0, r, t, 0.0)

    assert norm(x) > tol
    assert norm(r) / norm(b) < tol

def test_prec_2D(arpack_eigs, interface, x, num_evs, tol, atol, interactive=False):
    from transiflow.interface import JaDaHYMLS

    from jadapy import EpetraInterface
    from jadapy import jdqz

    numpy.random.seed(1234)

    jac_op = EpetraInterface.CrsMatrix(interface.jacobian(x))
    mass_op = EpetraInterface.CrsMatrix(interface.mass_matrix())
    jada_interface = JaDaHYMLS.Interface(interface, preconditioned_solve=False)

    alpha, beta = jdqz.jdqz(jac_op, mass_op, num_evs, tol=tol, subspace_dimensions=[20, 40],
                            interface=jada_interface, prec=jada_interface.prec)
    jdqz_eigs = numpy.array(sorted(alpha / beta, key=lambda x: abs(x)))

    assert_allclose(jdqz_eigs.real, arpack_eigs.real, rtol=0, atol=atol)
    assert_allclose(abs(jdqz_eigs.imag), abs(arpack_eigs.imag), rtol=0, atol=atol)

    if not interactive:
        return

    fig, ax = plt.subplots()
    ax.scatter(jdqz_eigs.real, jdqz_eigs.imag, marker='+')
    plt.show()

def test_prec_solve_2D(arpack_eigs, interface, x, num_evs, tol, atol, interactive=False):
    from transiflow.interface import JaDaHYMLS

    from jadapy import EpetraInterface
    from jadapy import jdqz

    numpy.random.seed(1234)

    jac_op = EpetraInterface.CrsMatrix(interface.jacobian(x))
    mass_op = EpetraInterface.CrsMatrix(interface.mass_matrix())
    jada_interface = JaDaHYMLS.Interface(interface, preconditioned_solve=True)

    alpha, beta = jdqz.jdqz(jac_op, mass_op, num_evs, tol=tol, subspace_dimensions=[20, 40],
                            interface=jada_interface)
    jdqz_eigs = numpy.array(sorted(alpha / beta, key=lambda x: abs(x)))

    assert_allclose(jdqz_eigs.real, arpack_eigs.real, rtol=0, atol=atol)
    assert_allclose(abs(jdqz_eigs.imag), abs(arpack_eigs.imag), rtol=0, atol=atol)

    if not interactive:
        return

    fig, ax = plt.subplots()
    ax.scatter(jdqz_eigs.real, jdqz_eigs.imag, marker='+')
    plt.show()

def test_complex_prec_2D(arpack_eigs, interface, x, num_evs, tol, atol, interactive=False):
    from transiflow.interface import JaDaHYMLS

    from jadapy import ComplexEpetraInterface
    from jadapy import jdqz

    numpy.random.seed(1234)

    jac_op = ComplexEpetraInterface.CrsMatrix(interface.jacobian(x))
    mass_op = ComplexEpetraInterface.CrsMatrix(interface.mass_matrix())
    jada_interface = JaDaHYMLS.ComplexInterface(interface, preconditioned_solve=False)

    alpha, beta = jdqz.jdqz(jac_op, mass_op, num_evs, tol=tol, subspace_dimensions=[20, 40],
                            interface=jada_interface, prec=jada_interface.prec)
    jdqz_eigs = numpy.array(sorted(alpha / beta, key=lambda x: abs(x)))

    assert_allclose(jdqz_eigs.real, arpack_eigs.real, rtol=0, atol=atol)
    assert_allclose(abs(jdqz_eigs.imag), abs(arpack_eigs.imag), rtol=0, atol=atol)

    if not interactive:
        return

    fig, ax = plt.subplots()
    ax.scatter(jdqz_eigs.real, jdqz_eigs.imag, marker='+')
    plt.show()

def test_complex_prec_solve_2D(arpack_eigs, interface, x, num_evs, tol, atol, interactive=False):
    from transiflow.interface import JaDaHYMLS

    from jadapy import ComplexEpetraInterface
    from jadapy import jdqz

    numpy.random.seed(1234)

    jac_op = ComplexEpetraInterface.CrsMatrix(interface.jacobian(x))
    mass_op = ComplexEpetraInterface.CrsMatrix(interface.mass_matrix())
    jada_interface = JaDaHYMLS.ComplexInterface(interface, preconditioned_solve=True)

    alpha, beta = jdqz.jdqz(jac_op, mass_op, num_evs, tol=tol, subspace_dimensions=[20, 40],
                            interface=jada_interface)
    jdqz_eigs = numpy.array(sorted(alpha / beta, key=lambda x: abs(x)))

    assert_allclose(jdqz_eigs.real, arpack_eigs.real, rtol=0, atol=atol)
    assert_allclose(abs(jdqz_eigs.imag), abs(arpack_eigs.imag), rtol=0, atol=atol)

    if not interactive:
        return

    fig, ax = plt.subplots()
    ax.scatter(jdqz_eigs.real, jdqz_eigs.imag, marker='+')
    plt.show()

def test_bordered_prec_solve_2D(arpack_eigs, interface, x, num_evs, tol, atol, interactive=False):
    from transiflow.interface import JaDaHYMLS

    from jadapy import EpetraInterface
    from jadapy import jdqz

    numpy.random.seed(1234)

    jac_op = EpetraInterface.CrsMatrix(interface.jacobian(x))
    mass_op = EpetraInterface.CrsMatrix(interface.mass_matrix())
    jada_interface = JaDaHYMLS.BorderedInterface(interface)

    alpha, beta = jdqz.jdqz(jac_op, mass_op, num_evs, tol=tol, subspace_dimensions=[20, 40],
                            interface=jada_interface)
    jdqz_eigs = numpy.array(sorted(alpha / beta, key=lambda x: abs(x)))

    assert_allclose(jdqz_eigs.real, arpack_eigs.real, rtol=0, atol=atol)
    assert_allclose(abs(jdqz_eigs.imag), abs(arpack_eigs.imag), rtol=0, atol=atol)

    if not interactive:
        return

    fig, ax = plt.subplots()
    ax.scatter(jdqz_eigs.real, jdqz_eigs.imag, marker='+')
    plt.show()
